################################################################################
#                      skforecast.deep_learning.utils                          #
#                                                                              #
# This work by skforecast team is licensed under the BSD 3-Clause License      #
################################################################################
# coding=utf-8

from __future__ import annotations
from typing import Any
from copy import deepcopy
import numpy as np
import pandas as pd
from ..utils import (
    initialize_lags,
    input_to_frame,
    check_optional_dependency
)

try:
    import keras
    from keras.layers import (
        Input,
        LSTM,
        GRU,
        SimpleRNN,
        RepeatVector,
        Concatenate,
        Dense,
        Reshape,
        TimeDistributed,
    )
    from keras.optimizers import Adam
    from keras.losses import MeanSquaredError
    from keras.models import Model
except ImportError as e:
    import sys
    if sys.version_info >= (3, 13):
        raise ImportError(
            "Python 3.13 is not supported by TensorFlow, which is the default "
            "backend used by Keras. To use Keras with Python 3.13, the KERAS_BACKEND "
            "environment variable needs to be set to 'torch', `os.environ['KERAS_BACKEND'] = 'torch'`."
            "Make sure you have PyTorch installed to use Keras with the torch backend. "
            "For installation instructions, visit https://pytorch.org/get-started/locally/"
        )
    else:
        package_name = str(e).split(" ")[-1].replace("'", "")
        check_optional_dependency(package_name=package_name)


def create_and_compile_model(
    series: pd.DataFrame,
    lags: int | list[int] | np.ndarray[int] | range[int],
    steps: int,
    levels: str | list[str] | tuple[str] | None = None,
    exog: pd.Series | pd.DataFrame | None = None,
    recurrent_layer: str = "LSTM",
    recurrent_units: int | list[int] | tuple[int] = 100,
    recurrent_layers_kwargs: dict[str, Any] | list[dict[str, Any]] | None = {"activation": "tanh"},
    dense_units: int | list[int] | tuple[int] | None = 64,
    dense_layers_kwargs: dict[str, Any] | list[dict[str, Any]] | None = {"activation": "relu"},
    output_dense_layer_kwargs: dict[str, Any] | None = {"activation": "linear"},
    compile_kwargs: dict[str, Any] = {"optimizer": Adam(), "loss": MeanSquaredError()},
    model_name: str | None = None
) -> keras.models.Model:
    """
    Build and compile a RNN-based Keras model for time series prediction, 
    supporting exogenous variables.

    Parameters
    ----------
    series : pandas DataFrame
        Input time series with shape (n_obs, n_series). Each column is a time series.
    lags : int, list, numpy ndarray, range
        Number of lagged time steps to consider in the input, index starts at 1, 
        so lag 1 is equal to t-1.
    
        - `int`: include lags from 1 to `lags` (included).
        - `list`, `1d numpy ndarray` or `range`: include only lags present in 
        `lags`, all elements must be int.
    steps : int
        Number of steps to predict.
    levels : str, list, default None
        Output level(s) (features) to predict. If None, defaults to the names of 
        input series.
    exog : pandas Series, pandas DataFrame, default None
        Exogenous variables to be included as input, should have the same number 
        of rows as `series`.
    recurrent_layer : str, default 'LSTM'
        Type of recurrent layer to be used, 'LSTM' [1]_, 'GRU' [2]_, or 'RNN' [3]_.
    recurrent_units : int, list, default 100
        Number of units in the recurrent layer(s). Can be an integer for single 
        recurrent layer, or a list of integers for multiple recurrent layers.
    recurrent_layers_kwargs : dict, list, default {'activation': 'tanh'}
        Additional keyword arguments for the recurrent layers [1]_, [2]_, [3]_. 
        Can be a single dictionary for all layers or a list of dictionaries 
        specifying different parameters for each recurrent layer.
    dense_units : int, list, tuple, None, default 64
        Number of units in the dense layer(s) [4]_. Can be an integer for single
        dense layer, or a list of integers for multiple dense layers.
    dense_layers_kwargs : dict, list, default {'activation': 'relu'}
        Additional keyword arguments for the dense layers [4]_. Can be a single
        dictionary for all layers or a list of dictionaries specifying different
        parameters for each dense layer.
    output_dense_layer_kwargs : dict, default {'activation': 'linear'}
        Additional keyword arguments for the output dense layer.
    compile_kwargs : dict, default {'optimizer': Adam(), 'loss': MeanSquaredError()}
        Additional keyword arguments for the model compilation, such as optimizer 
        and loss function. [5]_
    model_name : str, default None
        Name of the model.

    Returns
    -------
    model : keras.models.Model
        Compiled Keras model ready for training.

    References
    ----------
    .. [1] LSTM layer Keras documentation.
           https://keras.io/api/layers/recurrent_layers/lstm/

    .. [2] GRU layer Keras documentation.
           https://keras.io/api/layers/recurrent_layers/gru/

    .. [3] SimpleRNN layer Keras documentation.
           https://keras.io/api/layers/recurrent_layers/simple_rnn/

    .. [4] Dense layer Keras documentation.
           https://keras.io/api/layers/core_layers/dense/

    .. [5] Model training APIs: compile method.
           https://keras.io/api/models/model_training_apis/
    
    """

    keras_backend = keras.backend.backend()

    print(f"keras version: {keras.__version__}")
    print(f"Using backend: {keras_backend}")
    if keras_backend == "tensorflow":
        import tensorflow
        print(f"tensorflow version: {tensorflow.__version__}")
    elif keras_backend == "torch":
        import torch
        print(f"torch version: {torch.__version__}")
    elif keras_backend == "jax":
        import jax
        print(f"jax version: {jax.__version__}")
    else:
        print("Backend not recognized")
    print("")

    if exog is None:
        model = _create_and_compile_model_no_exog(
            series=series,
            lags=lags,
            steps=steps,
            levels=levels,
            recurrent_layer=recurrent_layer,
            recurrent_units=recurrent_units,
            recurrent_layers_kwargs=recurrent_layers_kwargs,
            dense_units=dense_units,
            dense_layers_kwargs=dense_layers_kwargs,
            output_dense_layer_kwargs=output_dense_layer_kwargs,
            compile_kwargs=compile_kwargs,
            model_name=model_name
        )
    else:
        model = _create_and_compile_model_exog(
            series=series,
            lags=lags,
            steps=steps,
            levels=levels,
            exog=exog,
            recurrent_layer=recurrent_layer,
            recurrent_units=recurrent_units,
            recurrent_layers_kwargs=recurrent_layers_kwargs,
            dense_units=dense_units,
            dense_layers_kwargs=dense_layers_kwargs,
            output_dense_layer_kwargs=output_dense_layer_kwargs,
            compile_kwargs=compile_kwargs,
            model_name=model_name
        )
    
    return model


def _create_and_compile_model_exog(
    series: pd.DataFrame,
    lags: int | list[int] | np.ndarray[int] | range[int],
    steps: int,
    levels: str | list[str] | tuple[str] | None = None,
    exog: pd.Series | pd.DataFrame | None = None,
    recurrent_layer: str = "LSTM",
    recurrent_units: int | list[int] | tuple[int] = 100,
    recurrent_layers_kwargs: dict[str, Any] | list[dict[str, Any]] | None = {"activation": "tanh"},
    dense_units: int | list[int] | tuple[int] | None = 64,
    dense_layers_kwargs: dict[str, Any] | list[dict[str, Any]] | None = {"activation": "relu"},
    output_dense_layer_kwargs: dict[str, Any] | None = {"activation": "linear"},
    compile_kwargs: dict[str, Any] = {"optimizer": Adam(), "loss": MeanSquaredError()},
    model_name: str | None = None
) -> keras.models.Model:
    """
    Build and compile a RNN-based Keras model for time series prediction, 
    supporting exogenous variables.

    Parameters
    ----------
    series : pandas DataFrame
        Input time series with shape (n_obs, n_series). Each column is a time series.
    lags : int, list, numpy ndarray, range
        Number of lagged time steps to consider in the input, index starts at 1, 
        so lag 1 is equal to t-1.
    
        - `int`: include lags from 1 to `lags` (included).
        - `list`, `1d numpy ndarray` or `range`: include only lags present in 
        `lags`, all elements must be int.
    steps : int
        Number of steps to predict.
    levels : str, list, default None
        Output level(s) (features) to predict. If None, defaults to the names of 
        input series.
    exog : pandas Series, pandas DataFrame, default None
        Exogenous variables to be included as input, should have the same number 
        of rows as `series`.
    recurrent_layer : str, default 'LSTM'
        Type of recurrent layer to be used, 'LSTM' [1]_, 'GRU' [2]_, or 'RNN' [3]_.
    recurrent_units : int, list, default 100
        Number of units in the recurrent layer(s). Can be an integer for single 
        recurrent layer, or a list of integers for multiple recurrent layers.
    recurrent_layers_kwargs : dict, list, default {'activation': 'tanh'}
        Additional keyword arguments for the recurrent layers [1]_, [2]_, [3]_. 
        Can be a single dictionary for all layers or a list of dictionaries 
        specifying different parameters for each recurrent layer.
    dense_units : int, list, tuple, None, default 64
        Number of units in the dense layer(s) [4]_. Can be an integer for single
        dense layer, or a list of integers for multiple dense layers.
    dense_layers_kwargs : dict, list, default {'activation': 'relu'}
        Additional keyword arguments for the dense layers [4]_. Can be a single
        dictionary for all layers or a list of dictionaries specifying different
        parameters for each dense layer.
    output_dense_layer_kwargs : dict, default {'activation': 'linear'}
        Additional keyword arguments for the output dense layer.
    compile_kwargs : dict, default {'optimizer': Adam(), 'loss': MeanSquaredError()}
        Additional keyword arguments for the model compilation, such as optimizer 
        and loss function. [5]_
    model_name : str, default None
        Name of the model.

    Returns
    -------
    model : keras.models.Model
        Compiled Keras model ready for training.

    References
    ----------
    .. [1] LSTM layer Keras documentation.
           https://keras.io/api/layers/recurrent_layers/lstm/

    .. [2] GRU layer Keras documentation.
           https://keras.io/api/layers/recurrent_layers/gru/

    .. [3] SimpleRNN layer Keras documentation.
           https://keras.io/api/layers/recurrent_layers/simple_rnn/

    .. [4] Dense layer Keras documentation.
           https://keras.io/api/layers/core_layers/dense/

    .. [5] Model training APIs: compile method.
           https://keras.io/api/models/model_training_apis/

    """

    if not isinstance(series, pd.DataFrame):
        raise TypeError(
            f"`series` must be a pandas DataFrame. Got {type(series)}."
        )
    n_series = series.shape[1]

    if exog is not None: 
        if not isinstance(exog, (pd.Series, pd.DataFrame)):
            raise TypeError(
                f"`exog` must be a pandas Series, DataFrame or None. Got {type(exog)}."
            )
        exog = input_to_frame(data=exog, input_name='exog')
        n_exog = exog.shape[1]
    else:
        n_exog = 0

    lags, _, _ = initialize_lags('ForecasterRNN', lags)
    n_lags = len(lags)

    if not isinstance(steps, int):
        raise TypeError(
            f"`steps` argument must be an int greater than or equal to 1. "
            f"Got {type(steps)}."
        )

    if steps < 1:
        raise ValueError(
            f"`steps` argument must be greater than or equal to 1. Got {steps}."
        )

    if levels is None:
        n_levels = n_series
    else:
        if isinstance(levels, str):
            levels = [levels]
        elif not isinstance(levels, (list, tuple)):
            raise TypeError(f"Invalid type for `levels`: {type(levels)}.")
        
        series_names_in = series.columns.tolist()
        missing_levels = [level for level in levels if level not in series_names_in]
        if missing_levels:
            raise ValueError(
                f"Levels {missing_levels} not found in series columns: {series_names_in}."
            )
        
        n_levels = len(levels)

    series_input = Input(shape=(n_lags, n_series), name="series_input")
    inputs = [series_input]
    if exog is not None:
        exog_input = Input(shape=(steps, n_exog), name="exog_input")
        inputs.append(exog_input)

    x = series_input
    if not isinstance(recurrent_units, (list, tuple, int)):
        raise TypeError(
            f"At least one recurrent layer (LSTM, GRU, or SimpleRNN) is required."
            f"`recurrent_units` argument must be an int or a list of ints. "
            f"Got {type(recurrent_units)}."
        )
    if not isinstance(recurrent_units, (list, tuple)):
        recurrent_units = [recurrent_units]

    if isinstance(recurrent_layers_kwargs, dict):
        recurrent_layers_kwargs = [recurrent_layers_kwargs] * len(recurrent_units)
    elif isinstance(recurrent_layers_kwargs, (list, tuple)):
        if len(recurrent_layers_kwargs) != len(recurrent_units):
            raise ValueError(
                "If `recurrent_layers_kwargs` is a list, it must have the same "
                "length as `recurrent_units`. One dict of kwargs per recurrent layer."
            )
    elif recurrent_layers_kwargs is None:
        recurrent_layers_kwargs = [{}] * len(recurrent_units)
    else:
        raise TypeError(
            f"`recurrent_layers_kwargs` must be a dict, a list of dicts or None. "
            f"Got {type(recurrent_layers_kwargs)}."
        )

    for i, units in enumerate(recurrent_units):

        return_sequences = i < len(recurrent_units) - 1

        layer_kwargs = deepcopy(recurrent_layers_kwargs[i])
        layer_kwargs.update({
            "units": units,
            "return_sequences": return_sequences,
        })
        if "name" not in layer_kwargs:
            layer_kwargs["name"] = f"{recurrent_layer.lower()}_{i + 1}"
        
        if recurrent_layer == "LSTM":
            x = LSTM(**layer_kwargs)(x)
        elif recurrent_layer == "GRU":
            x = GRU(**layer_kwargs)(x)
        elif recurrent_layer == "RNN":
            x = SimpleRNN(**layer_kwargs)(x)
        else:
            valid_layers = ["LSTM", "GRU", "RNN"]
            raise ValueError(
                f"`recurrent_layer` must be one of {valid_layers}. Got '{recurrent_layer}'."
            )

    # NOTE: Shape (batch, steps, features)
    x = RepeatVector(steps, name="repeat_vector")(x)

    if exog is not None:
        # NOTE: Shape (batch, steps, features + n_exog)
        x = Concatenate(axis=-1, name="concat_exog")([x, exog_input])

    if dense_units is not None:
        if not isinstance(dense_units, (list, tuple, int)):
            raise TypeError(
                f"`dense_units` argument must be an int, a list of ints or None. "
                f"Got {type(dense_units)}."
            )
        if not isinstance(dense_units, (list, tuple)):
            dense_units = [dense_units]

        if isinstance(dense_layers_kwargs, dict):
            dense_layers_kwargs = [dense_layers_kwargs] * len(dense_units)
        elif isinstance(dense_layers_kwargs, (list, tuple)):
            if len(dense_layers_kwargs) != len(dense_units):
                raise ValueError(
                    "If `dense_layers_kwargs` is a list, it must have the same "
                    "length as `dense_units`. One dict of kwargs per dense layer."
                )
        elif dense_layers_kwargs is None:
            dense_layers_kwargs = [{}] * len(dense_units)
        else:
            raise TypeError(
                f"`dense_layers_kwargs` must be a dict, a list of dicts or None. "
                f"Got {type(dense_layers_kwargs)}."
            )
        
        for i, units in enumerate(dense_units):
            
            layer_kwargs = deepcopy(dense_layers_kwargs[i])
            layer_kwargs.update({
                "units": units,
            })
            if "name" in layer_kwargs:
                layer_name = layer_kwargs.pop("name")
            else:
                layer_name = f"dense_td_{i + 1}"

            x = TimeDistributed(Dense(**layer_kwargs), name=layer_name)(x)

    if output_dense_layer_kwargs is None:
        output_layer_kwargs = {}
    else:
        output_layer_kwargs = deepcopy(output_dense_layer_kwargs)
    
    output_layer_kwargs.update({
        "units": n_levels,
    })
    if "name" in output_layer_kwargs:
        layer_name = output_layer_kwargs.pop("name")
    else:
        layer_name = "output_dense_td_layer"

    output = TimeDistributed(Dense(**output_layer_kwargs), name=layer_name)(x)

    model = Model(inputs=inputs, outputs=output, name=model_name)
    model.compile(**compile_kwargs)

    return model


def _create_and_compile_model_no_exog(
    series: pd.DataFrame,
    lags: int | list[int] | np.ndarray[int] | range[int],
    steps: int,
    levels: str | list[str] | tuple[str] | None = None,
    recurrent_layer: str = "LSTM",
    recurrent_units: int | list[int] | tuple[int] = 100,
    recurrent_layers_kwargs: dict[str, Any] | list[dict[str, Any]] | None = {"activation": "tanh"},
    dense_units: int | list[int] | tuple[int] | None = 64,
    dense_layers_kwargs: dict[str, Any] | list[dict[str, Any]] | None = {"activation": "relu"},
    output_dense_layer_kwargs: dict[str, Any] | None = {"activation": "linear"},
    compile_kwargs: dict[str, Any] = {"optimizer": Adam(), "loss": MeanSquaredError()},
    model_name: str | None = None
) -> keras.models.Model:
    """
    Build and compile a RNN-based Keras model for time series prediction.

    Parameters
    ----------
    series : pandas DataFrame
        Input time series with shape (n_obs, n_series). Each column is a time series.
    lags : int, list, numpy ndarray, range
        Number of lagged time steps to consider in the input, index starts at 1, 
        so lag 1 is equal to t-1.
    
        - `int`: include lags from 1 to `lags` (included).
        - `list`, `1d numpy ndarray` or `range`: include only lags present in 
        `lags`, all elements must be int.
    steps : int
        Number of steps to predict.
    levels : str, list, default None
        Output level(s) (features) to predict. If None, defaults to the names of 
        input series.
    recurrent_layer : str, default 'LSTM'
        Type of recurrent layer to be used, 'LSTM' [1]_, 'GRU' [2]_, or 'RNN' [3]_.
    recurrent_units : int, list, default 100
        Number of units in the recurrent layer(s). Can be an integer for single 
        recurrent layer, or a list of integers for multiple recurrent layers.
    recurrent_layers_kwargs : dict, list, default {'activation': 'tanh'}
        Additional keyword arguments for the recurrent layers [1]_, [2]_, [3]_. 
        Can be a single dictionary for all layers or a list of dictionaries 
        specifying different parameters for each recurrent layer.
    dense_units : int, list, tuple, None, default 64
        Number of units in the dense layer(s) [4]_. Can be an integer for single
        dense layer, or a list of integers for multiple dense layers.
    dense_layers_kwargs : dict, list, default {'activation': 'relu'}
        Additional keyword arguments for the dense layers [4]_. Can be a single
        dictionary for all layers or a list of dictionaries specifying different
        parameters for each dense layer.
    output_dense_layer_kwargs : dict, default {'activation': 'linear'}
        Additional keyword arguments for the output dense layer.
    compile_kwargs : dict, default {'optimizer': Adam(), 'loss': MeanSquaredError()}
        Additional keyword arguments for the model compilation, such as optimizer 
        and loss function. [5]_
    model_name : str, default None
        Name of the model.

    Returns
    -------
    model : keras.models.Model
        Compiled Keras model ready for training.

    References
    ----------
    .. [1] LSTM layer Keras documentation.
           https://keras.io/api/layers/recurrent_layers/lstm/

    .. [2] GRU layer Keras documentation.
           https://keras.io/api/layers/recurrent_layers/gru/

    .. [3] SimpleRNN layer Keras documentation.
           https://keras.io/api/layers/recurrent_layers/simple_rnn/

    .. [4] Dense layer Keras documentation.
           https://keras.io/api/layers/core_layers/dense/

    .. [5] Model training APIs: compile method.
           https://keras.io/api/models/model_training_apis/

    """

    if not isinstance(series, pd.DataFrame):
        raise TypeError(
            f"`series` must be a pandas DataFrame. Got {type(series)}."
        )
    n_series = series.shape[1]

    lags, _, _ = initialize_lags('ForecasterRNN', lags)
    n_lags = len(lags)

    if not isinstance(steps, int):
        raise TypeError(
            f"`steps` argument must be an int greater than or equal to 1. "
            f"Got {type(steps)}."
        )

    if steps < 1:
        raise ValueError(
            f"`steps` argument must be greater than or equal to 1. Got {steps}."
        )

    if levels is None:
        n_levels = n_series
    else:
        if isinstance(levels, str):
            levels = [levels]
        elif not isinstance(levels, (list, tuple)):
            raise TypeError(f"Invalid type for `levels`: {type(levels)}.")
        
        series_names_in = series.columns.tolist()
        missing_levels = [level for level in levels if level not in series_names_in]
        if missing_levels:
            raise ValueError(
                f"Levels {missing_levels} not found in series columns: {series_names_in}."
            )
        
        n_levels = len(levels)

    series_input = Input(shape=(n_lags, n_series), name="series_input")
    x = series_input

    # Recurrent units
    if not isinstance(recurrent_units, (list, tuple, int)):
        raise TypeError(
            f"At least one recurrent layer (LSTM, GRU, or SimpleRNN) is required."
            f"`recurrent_units` argument must be an int or a list of ints. "
            f"Got {type(recurrent_units)}."
        )
    if not isinstance(recurrent_units, (list, tuple)):
        recurrent_units = [recurrent_units]
    
    if isinstance(recurrent_layers_kwargs, dict):
        recurrent_layers_kwargs = [recurrent_layers_kwargs] * len(recurrent_units)
    elif isinstance(recurrent_layers_kwargs, (list, tuple)):
        if len(recurrent_layers_kwargs) != len(recurrent_units):
            raise ValueError(
                "If `recurrent_layers_kwargs` is a list, it must have the same "
                "length as `recurrent_units`. One dict of kwargs per recurrent layer."
            )
    elif recurrent_layers_kwargs is None:
        recurrent_layers_kwargs = [{}] * len(recurrent_units)
    else:
        raise TypeError(
            f"`recurrent_layers_kwargs` must be a dict, a list of dicts or None. "
            f"Got {type(recurrent_layers_kwargs)}."
        )

    for i, units in enumerate(recurrent_units):

        return_sequences = i < len(recurrent_units) - 1

        layer_kwargs = deepcopy(recurrent_layers_kwargs[i])
        layer_kwargs.update({
            "units": units,
            "return_sequences": return_sequences,
        })
        if "name" not in layer_kwargs:
            layer_kwargs["name"] = f"{recurrent_layer.lower()}_{i + 1}"
        
        if recurrent_layer == "LSTM":
            x = LSTM(**layer_kwargs)(x)
        elif recurrent_layer == "GRU":
            x = GRU(**layer_kwargs)(x)
        elif recurrent_layer == "RNN":
            x = SimpleRNN(**layer_kwargs)(x)
        else:
            valid_layers = ["LSTM", "GRU", "RNN"]
            raise ValueError(
                f"`recurrent_layer` must be one of {valid_layers}. Got '{recurrent_layer}'."
            )

    # Dense layers
    if dense_units is not None:
        if not isinstance(dense_units, (list, tuple, int)):
            raise TypeError(
                f"`dense_units` argument must be an int, a list of ints or None. "
                f"Got {type(dense_units)}."
            )
        if not isinstance(dense_units, (list, tuple)):
            dense_units = [dense_units]

        if isinstance(dense_layers_kwargs, dict):
            dense_layers_kwargs = [dense_layers_kwargs] * len(dense_units)
        elif isinstance(dense_layers_kwargs, (list, tuple)):
            if len(dense_layers_kwargs) != len(dense_units):
                raise ValueError(
                    "If `dense_layers_kwargs` is a list, it must have the same "
                    "length as `dense_units`. One dict of kwargs per dense layer."
                )
        elif dense_layers_kwargs is None:
            dense_layers_kwargs = [{}] * len(dense_units)
        else:
            raise TypeError(
                f"`dense_layers_kwargs` must be a dict, a list of dicts or None. "
                f"Got {type(dense_layers_kwargs)}."
            )
        
        for i, units in enumerate(dense_units):
        
            layer_kwargs = deepcopy(dense_layers_kwargs[i])
            layer_kwargs.update({
                "units": units,
            })
            if "name" not in layer_kwargs:
                layer_kwargs['name'] = f"dense_{i + 1}"
            
            x = Dense(**layer_kwargs)(x)

    # Output layer
    if output_dense_layer_kwargs is None:
        output_layer_kwargs = {}
    else:
        output_layer_kwargs = deepcopy(output_dense_layer_kwargs)
    
    output_layer_kwargs.update({
        "units": n_levels * steps,
    })
    if "name" not in output_layer_kwargs:
        output_layer_kwargs['name'] = "output_dense_td_layer"
    
    x = Dense(**output_layer_kwargs)(x)
    output_layer = Reshape((steps, n_levels), name="reshape")(x)

    model = Model(inputs=series_input, outputs=output_layer, name=model_name)
    model.compile(**compile_kwargs)

    return model



################################################################################
#                       skforecast.feature_selection                           #
#                                                                              #
# This work by skforecast team is licensed under the BSD 3-Clause License.     #
################################################################################
# coding=utf-8

from __future__ import annotations
import re
from copy import deepcopy
from itertools import chain
import warnings
import numpy as np
import pandas as pd


def select_features(
    forecaster: object,
    selector: object,
    y: pd.Series | pd.DataFrame,
    exog: pd.Series | pd.DataFrame | None = None,
    select_only: str | None = None,
    force_inclusion: list[str] | str | None = None,
    subsample: int | float = 0.5,
    random_state: int = 123,
    verbose: bool = True
) -> tuple[list[int], list[str], list[str]]:
    """
    Feature selection using any of the sklearn.feature_selection module selectors 
    (such as `RFECV`, `SelectFromModel`, etc.). Two groups of features are
    evaluated: autoregressive features (lags and window features) and exogenous
    features. By default, the selection process is performed on both sets of features
    at the same time, so that the most relevant autoregressive and exogenous features
    are selected. However, using the `select_only` argument, the selection process
    can focus only on the autoregressive or exogenous features without taking into
    account the other features. Therefore, all other features will remain in the model. 
    It is also possible to force the inclusion of certain features in the final list
    of selected features using the `force_inclusion` parameter.

    Parameters
    ----------
    forecaster : ForecasterRecursive, ForecasterDirect
        Forecaster model. If forecaster is a ForecasterDirect, the
        selector will only be applied to the features of the first step.
    selector : object
        A feature selector from sklearn.feature_selection.
    y : pandas Series, pandas DataFrame
        Target time series to which the feature selection will be applied.
    exog : pandas Series, pandas DataFrame, default None
        Exogenous variable/s included as predictor/s. Must have the same
        number of observations as `y` and should be aligned so that y[i] is
        regressed on exog[i].
    select_only : str, default None
        Decide what type of features to include in the selection process. 
        
        - If `'autoreg'`, only autoregressive features (lags and window features)
        are evaluated by the selector. All exogenous features are included in the
        output `selected_exog`.
        - If `'exog'`, only exogenous features are evaluated without the presence
        of autoregressive features. All autoregressive features are included 
        in the outputs `selected_lags` and `selected_window_features`.
        - If `None`, all features are evaluated by the selector.
    force_inclusion : list, str, default None
        Features to force include in the final list of selected features.
        
        - If `list`, list of feature names to force include.
        - If `str`, regular expression to identify features to force include. 
        For example, if `force_inclusion="^sun_"`, all features that begin 
        with "sun_" will be included in the final list of selected features.
    subsample : int, float, default 0.5
        Proportion of records to use for feature selection.
    random_state : int, default 123
        Sets a seed for the random subsample so that the subsampling process 
        is always deterministic.
    verbose : bool, default True
        Print information about feature selection process.

    Returns
    -------
    selected_lags : list
        List of selected lags.
    selected_window_features : list
        List of selected window features.
    selected_exog : list
        List of selected exogenous features.

    """

    forecaster_name = type(forecaster).__name__
    valid_forecasters = ['ForecasterRecursive', 'ForecasterDirect']

    if forecaster_name not in valid_forecasters:
        raise TypeError(
            f"`forecaster` must be one of the following classes: {valid_forecasters}."
        )
    
    if select_only not in ['autoreg', 'exog', None]:
        raise ValueError(
            "`select_only` must be one of the following values: 'autoreg', 'exog', None."
        )

    if subsample <= 0 or subsample > 1:
        raise ValueError(
            "`subsample` must be a number greater than 0 and less than or equal to 1."
        )
    
    forecaster = deepcopy(forecaster)
    forecaster.is_fitted = False
    X_train, y_train = forecaster.create_train_X_y(y=y, exog=exog)
    if forecaster_name == 'ForecasterDirect':
        X_train, y_train = forecaster.filter_train_X_y_for_step(
                               step          = 1,
                               X_train       = X_train,
                               y_train       = y_train,
                               remove_suffix = True
                           )
    
    lags_cols = []
    window_features_cols = []
    autoreg_cols = []
    if forecaster.lags is not None:
        lags_cols = forecaster.lags_names
        autoreg_cols.extend(lags_cols)
    if forecaster.window_features is not None:
        window_features_cols = forecaster.window_features_names
        autoreg_cols.extend(window_features_cols)

    exog_cols = [col for col in X_train.columns if col not in autoreg_cols]

    forced_autoreg = []
    forced_exog = []
    if force_inclusion is not None:
        if isinstance(force_inclusion, list):
            forced_autoreg = [col for col in force_inclusion if col in autoreg_cols]
            forced_exog = [col for col in force_inclusion if col in exog_cols]
        elif isinstance(force_inclusion, str):
            forced_autoreg = [col for col in autoreg_cols if re.match(force_inclusion, col)]
            forced_exog = [col for col in exog_cols if re.match(force_inclusion, col)]

    if select_only == 'autoreg':
        X_train = X_train.drop(columns=exog_cols)
    elif select_only == 'exog':
        X_train = X_train.drop(columns=autoreg_cols)

    if isinstance(subsample, float):
        subsample = int(len(X_train) * subsample)

    rng = np.random.default_rng(seed=random_state)
    sample = rng.integers(low=0, high=len(X_train), size=subsample)
    X_train_sample = X_train.iloc[sample, :]
    y_train_sample = y_train.iloc[sample]
    selector.fit(X_train_sample, y_train_sample)
    selected_features = selector.get_feature_names_out()

    if select_only == 'exog':
        selected_autoreg = autoreg_cols
    else:
        selected_autoreg = [
            feature
            for feature in selected_features
            if feature in autoreg_cols
        ]

    if select_only == 'autoreg':
        selected_exog = exog_cols
    else:
        selected_exog = [
            feature
            for feature in selected_features
            if feature in exog_cols
        ]

    if force_inclusion is not None: 
        if select_only != 'autoreg':
            forced_exog_not_selected = set(forced_exog) - set(selected_features)
            selected_exog.extend(forced_exog_not_selected)
            selected_exog.sort(key=exog_cols.index)
        if select_only != 'exog':
            forced_autoreg_not_selected = set(forced_autoreg) - set(selected_features)
            selected_autoreg.extend(forced_autoreg_not_selected)
            selected_autoreg.sort(key=autoreg_cols.index)

    if len(selected_autoreg) == 0:
        warnings.warn(
            "No autoregressive features have been selected. Since a Forecaster "
            "cannot be created without them, be sure to include at least one "
            "using the `force_inclusion` parameter."
        )
        selected_lags = []
        selected_window_features = []
    else:
        selected_lags = [
            int(feature.replace('lag_', '')) 
            for feature in selected_autoreg if feature in lags_cols
        ]
        selected_window_features = [
            feature for feature in selected_autoreg if feature in window_features_cols
        ]

    if verbose:
        print(f"Recursive feature elimination ({selector.__class__.__name__})")
        print("--------------------------------" + "-" * len(selector.__class__.__name__))
        print(f"Total number of records available: {X_train.shape[0]}")
        print(f"Total number of records used for feature selection: {X_train_sample.shape[0]}")
        print(f"Number of features available: {len(autoreg_cols) + len(exog_cols)}") 
        print(f"    Lags            (n={len(lags_cols)})")
        print(f"    Window features (n={len(window_features_cols)})")
        print(f"    Exog            (n={len(exog_cols)})")
        print(f"Number of features selected: {len(selected_features)}")
        print(f"    Lags            (n={len(selected_lags)}) : {selected_lags}")
        print(f"    Window features (n={len(selected_window_features)}) : {selected_window_features}")
        print(f"    Exog            (n={len(selected_exog)}) : {selected_exog}")

    return selected_lags, selected_window_features, selected_exog


def select_features_multiseries(
    forecaster: object,
    selector: object,
    series: pd.DataFrame | dict[str, pd.Series | pd.DataFrame],
    exog: pd.Series | pd.DataFrame | dict[str, pd.Series | pd.DataFrame] | None = None,
    select_only: str | None = None,
    force_inclusion: list[str] | str | None = None,
    subsample: int | float = 0.5,
    random_state: int = 123,
    verbose: bool = True,
) -> tuple[list[int] | dict[str, int], list[str], list[str]]:
    """
    Feature selection using any of the sklearn.feature_selection module selectors 
    (such as `RFECV`, `SelectFromModel`, etc.). Two groups of features are
    evaluated: autoregressive features and exogenous features. By default, the 
    selection process is performed on both sets of features at the same time, 
    so that the most relevant autoregressive and exogenous features are selected. 
    However, using the `select_only` argument, the selection process can focus 
    only on the autoregressive or exogenous features without taking into account 
    the other features. Therefore, all other features will remain in the model. 
    It is also possible to force the inclusion of certain features in the final 
    list of selected features using the `force_inclusion` parameter.

    Parameters
    ----------
    forecaster : ForecasterRecursiveMultiSeries, ForecasterDirectMultiVariate
        Forecaster model. If forecaster is a ForecasterDirectMultiVariate, the
        selector will only be applied to the features of the first step.
    selector : object
        A feature selector from sklearn.feature_selection.
    series : pandas DataFrame, dict
        Target time series to which the feature selection will be applied.
    exog : pandas Series, pandas DataFrame, dict, default None
        Exogenous variables.
    select_only : str, default None
        Decide what type of features to include in the selection process. 
        
        - If `'autoreg'`, only autoregressive features (lags and window features) 
        are evaluated by the selector. All exogenous features are 
        included in the output `selected_exog`.
        - If `'exog'`, only exogenous features are evaluated without the presence
        of autoregressive features. All autoregressive features are included 
        in the outputs `selected_lags` and `selected_window_features`.
        - If `None`, all features are evaluated by the selector.
    force_inclusion : list, str, default None
        Features to force include in the final list of selected features.
        
        - If `list`, list of feature names to force include.
        - If `str`, regular expression to identify features to force include. 
        For example, if `force_inclusion="^sun_"`, all features that begin 
        with "sun_" will be included in the final list of selected features.
    subsample : int, float, default 0.5
        Proportion of records to use for feature selection.
    random_state : int, default 123
        Sets a seed for the random subsample so that the subsampling process 
        is always deterministic.
    verbose : bool, default True
        Print information about feature selection process.

    Returns
    -------
    selected_lags : list, dict
        List of selected lags. If the forecaster is a ForecasterDirectMultiVariate,
        the output is a dict with the selected lags for each series, {series_name: lags},
        as the lags can be different for each series.
    selected_window_features : list
        List of selected window features.
    selected_exog : list
        List of selected exogenous features.

    """

    forecaster_name = type(forecaster).__name__
    valid_forecasters = [
        'ForecasterRecursiveMultiSeries',
        'ForecasterDirectMultiVariate'
    ]

    if forecaster_name not in valid_forecasters:
        raise TypeError(
            f"`forecaster` must be one of the following classes: {valid_forecasters}."
        )
    
    if select_only not in ['autoreg', 'exog', None]:
        raise ValueError(
            "`select_only` must be one of the following values: 'autoreg', 'exog', None."
        )

    if subsample <= 0 or subsample > 1:
        raise ValueError(
            "`subsample` must be a number greater than 0 and less than or equal to 1."
        )
    
    forecaster = deepcopy(forecaster)
    forecaster.is_fitted = False
    output = forecaster._create_train_X_y(series=series, exog=exog)
    X_train = output[0]
    y_train = output[1]
    if forecaster_name == 'ForecasterDirectMultiVariate':
        X_train, y_train = forecaster.filter_train_X_y_for_step(
                               step          = 1,
                               X_train       = X_train,
                               y_train       = y_train,
                               remove_suffix = True
                           )
        lags_cols = list(
            chain(*[v for v in forecaster.lags_names.values() if v is not None])
        )
        window_features_cols = forecaster.X_train_window_features_names_out_
        encoding_cols = []
    else:
        lags_cols = forecaster.lags_names
        window_features_cols = output[6]  # X_train_window_features_names_out_ output
        if forecaster.encoding == 'onehot':
            encoding_cols = output[4]  # X_train_series_names_in_ output
        else:
            encoding_cols = ['_level_skforecast']
    
    lags_cols = [] if lags_cols is None else lags_cols
    window_features_cols = [] if window_features_cols is None else window_features_cols
    autoreg_cols = []
    if forecaster.lags is not None:
        autoreg_cols.extend(lags_cols)
    if forecaster.window_features is not None:
        autoreg_cols.extend(window_features_cols)
    
    exog_cols = [
        col
        for col in X_train.columns
        if col not in autoreg_cols and col not in encoding_cols
    ]

    forced_autoreg = []
    forced_exog = []
    if force_inclusion is not None:
        if isinstance(force_inclusion, list):
            forced_autoreg = [col for col in force_inclusion if col in autoreg_cols]
            forced_exog = [col for col in force_inclusion if col in exog_cols]
        elif isinstance(force_inclusion, str):
            forced_autoreg = [col for col in autoreg_cols if re.match(force_inclusion, col)]
            forced_exog = [col for col in exog_cols if re.match(force_inclusion, col)]

    if select_only == 'autoreg':
        X_train = X_train.drop(columns=exog_cols + encoding_cols)
    elif select_only == 'exog':
        X_train = X_train.drop(columns=autoreg_cols + encoding_cols)
    else:
        X_train = X_train.drop(columns=encoding_cols)

    if isinstance(subsample, float):
        subsample = int(len(X_train) * subsample)

    rng = np.random.default_rng(seed=random_state)
    sample = rng.integers(low=0, high=len(X_train), size=subsample)
    X_train_sample = X_train.iloc[sample, :]
    y_train_sample = y_train.iloc[sample]
    selector.fit(X_train_sample, y_train_sample)
    selected_features = selector.get_feature_names_out()

    if select_only == 'exog':
        selected_autoreg = autoreg_cols
    else:
        selected_autoreg = [
            feature
            for feature in selected_features
            if feature in autoreg_cols
        ]

    if select_only == 'autoreg':
        selected_exog = exog_cols
    else:
        selected_exog = [
            feature
            for feature in selected_features
            if feature in exog_cols
        ]

    if force_inclusion is not None: 
        if select_only != 'autoreg':
            forced_exog_not_selected = set(forced_exog) - set(selected_features)
            selected_exog.extend(forced_exog_not_selected)
            selected_exog.sort(key=exog_cols.index)
        if select_only != 'exog':
            forced_autoreg_not_selected = set(forced_autoreg) - set(selected_features)
            selected_autoreg.extend(forced_autoreg_not_selected)
            selected_autoreg.sort(key=autoreg_cols.index)

    if len(selected_autoreg) == 0:
        warnings.warn(
            "No autoregressive features have been selected. Since a Forecaster "
            "cannot be created without them, be sure to include at least one "
            "using the `force_inclusion` parameter."
        )
        selected_lags = []
        selected_window_features = []
        verbose_selected_lags = []
    else:
        if forecaster_name == 'ForecasterDirectMultiVariate':
            selected_lags = {
                series_name: (
                    [
                        int(feature.replace(f"{series_name}_lag_", ""))
                        for feature in selected_autoreg
                        if feature in lags_names
                    ]
                    if lags_names is not None
                    else []
                )
                for series_name, lags_names in forecaster.lags_names.items()
            }
            verbose_selected_lags = [
                feature for feature in selected_autoreg if feature in lags_cols
            ]
        else:
            selected_lags = [
                int(feature.replace('lag_', '')) 
                for feature in selected_autoreg 
                if feature in lags_cols
            ]
            verbose_selected_lags = selected_lags

        selected_window_features = [
            feature for feature in selected_autoreg 
            if feature in window_features_cols
        ]

    if verbose:
        print(f"Recursive feature elimination ({selector.__class__.__name__})")
        print("--------------------------------" + "-" * len(selector.__class__.__name__))
        print(f"Total number of records available: {X_train.shape[0]}")
        print(f"Total number of records used for feature selection: {X_train_sample.shape[0]}")
        print(f"Number of features available: {len(autoreg_cols) + len(exog_cols)}") 
        print(f"    Lags            (n={len(lags_cols)})")
        print(f"    Window features (n={len(window_features_cols)})")
        print(f"    Exog            (n={len(exog_cols)})")
        print(f"Number of features selected: {len(selected_features)}")
        print(f"    Lags            (n={len(verbose_selected_lags)}) : {verbose_selected_lags}")
        print(f"    Window features (n={len(selected_window_features)}) : {selected_window_features}")
        print(f"    Exog            (n={len(selected_exog)}) : {selected_exog}")

    return selected_lags, selected_window_features, selected_exog



# Unit test initialize_differentiator_multiseries
# ==============================================================================
import re
import pytest
from skforecast.preprocessing import TimeSeriesDifferentiator
from skforecast.utils import initialize_differentiator_multiseries
from skforecast.exceptions import IgnoredArgumentWarning


def test_initialize_differentiator_when_differentiator_is_None():
    """
    Test initialize_differentiator_multiseries when `differentiator` is None.
    """
    
    differentiator_ = initialize_differentiator_multiseries(
                          series_names_in_ = ['col1', 'col2'],
                          differentiator   = None
                      )

    assert differentiator_ == {'col1': None, 'col2': None, '_unknown_level': None}


def test_initialize_differentiator_when_differentiator_is_TimeSeriesDifferentiator():
    """
    Test initialize_differentiator_multiseries when `differentiator` is a TimeSeriesDifferentiator.
    """
    series_names_in_ = ['col1', 'col2']
    differentiator = TimeSeriesDifferentiator(order=1, window_size=5)
    
    differentiator_ = initialize_differentiator_multiseries(
                          series_names_in_ = series_names_in_,
                          differentiator   = differentiator
                      )
    
    expected = {
        'col1': differentiator, 
        'col2': differentiator, 
        '_unknown_level': differentiator
    }

    assert differentiator_.keys() == expected.keys()
    for k in expected.keys():
        assert isinstance(differentiator_[k], TimeSeriesDifferentiator)
        assert differentiator_[k].order == expected[k].order
        assert differentiator_[k].window_size == expected[k].window_size


def test_initialize_differentiator_when_differentiator_is_dict():
    """
    Test initialize_differentiator_multiseries when `differentiator` is a dict.
    """
    series_names_in_ = ['col1', 'col2', 'col3']
    differentiator = {
        'col1': TimeSeriesDifferentiator(order=1, window_size=5),
        'col2': TimeSeriesDifferentiator(order=2, window_size=5),
        '_unknown_level': TimeSeriesDifferentiator(order=3, window_size=5)
    }
    
    differentiator_ = initialize_differentiator_multiseries(
                          series_names_in_ = series_names_in_,
                          differentiator   = differentiator
                      )
    
    expected = {
        'col1': TimeSeriesDifferentiator(order=1, window_size=5),
        'col2': TimeSeriesDifferentiator(order=2, window_size=5),
        'col3': None,
        '_unknown_level': TimeSeriesDifferentiator(order=3, window_size=5)
    }

    assert differentiator_.keys() == expected.keys()
    for k in expected.keys():
        if expected[k] is None:
            assert isinstance(differentiator_[k], type(None))
        else:
            assert isinstance(differentiator_[k], TimeSeriesDifferentiator)
            assert differentiator_[k].order == expected[k].order
            assert differentiator_[k].window_size == expected[k].window_size


def test_initialize_differentiator_multiseries_IgnoredArgumentWarning_when_keys_of_differentiator_not_equal_to_series_names_in_():
    """
    Test IgnoredArgumentWarning is raised when `differentiator` is a dict and its keys 
    are not the same as `series_names_in_`.
    """
    series_names_in_ = ['col1', 'col2']
    differentiator = {
        'col1': TimeSeriesDifferentiator(order=1, window_size=5),
        'col3': TimeSeriesDifferentiator(order=2, window_size=5),
        '_unknown_level': TimeSeriesDifferentiator(order=1, window_size=5)
    }
    
    series_not_in_differentiator = set(['col2'])
    warn_msg = re.escape(
        f"{series_not_in_differentiator} not present in `differentiation`."
        f" No differentiation is applied to these series."
    )
    with pytest.warns(IgnoredArgumentWarning, match = warn_msg):
        differentiator_ = initialize_differentiator_multiseries(
                              series_names_in_ = series_names_in_,
                              differentiator   = differentiator
                          )
    
    expected = {
        'col1': TimeSeriesDifferentiator(order=1, window_size=5),
        'col2': None,
        '_unknown_level': TimeSeriesDifferentiator(order=1, window_size=5)
    }

    assert differentiator_.keys() == expected.keys()
    for k in expected.keys():
        if expected[k] is None:
            assert isinstance(differentiator_[k], type(None))
        else:
            assert isinstance(differentiator_[k], TimeSeriesDifferentiator)
            assert differentiator_[k].order == expected[k].order
            assert differentiator_[k].window_size == expected[k].window_size



# Unit test _summary
# ============================================================================== 
from skforecast.drift_detection import RangeDriftDetector


def test_summary_no_out_of_range(capsys):
    """
    Test _summary when no out-of-range values are found.
    """
    out_of_range_series = []
    out_of_range_series_ranges = []
    out_of_range_exog = []
    out_of_range_exog_ranges = []

    RangeDriftDetector._summary(
        out_of_range_series=out_of_range_series,
        out_of_range_series_ranges=out_of_range_series_ranges,
        out_of_range_exog=out_of_range_exog,
        out_of_range_exog_ranges=out_of_range_exog_ranges
    )

    captured = capsys.readouterr()
    assert "No series with out-of-range values found." in captured.out
    assert "No exogenous variables with out-of-range values found." in captured.out


def test_summary_out_of_range_series(capsys):
    """
    Test _summary with out-of-range series.
    """
    out_of_range_series = ["series1", "series2"]
    out_of_range_series_ranges = [(1.0, 10.0), (2.0, 20.0)]
    out_of_range_exog = []
    out_of_range_exog_ranges = []

    RangeDriftDetector._summary(
        out_of_range_series=out_of_range_series,
        out_of_range_series_ranges=out_of_range_series_ranges,
        out_of_range_exog=out_of_range_exog,
        out_of_range_exog_ranges=out_of_range_exog_ranges
    )

    captured = capsys.readouterr()
    assert "'series1' has values outside the observed range [1.00000, 10.00000]." in captured.out
    assert "'series2' has values outside the observed range [2.00000, 20.00000]." in captured.out
    assert "No exogenous variables with out-of-range values found." in captured.out


def test_summary_out_of_range_exog(capsys):
    """
    Test _summary with out-of-range exogenous variables.
    """
    out_of_range_series = []
    out_of_range_series_ranges = []
    out_of_range_exog = {'series_1': ["exog1", "exog2"]}
    out_of_range_exog_ranges = {'series_1': [(0.0, 5.0), (10.0, 15.0)]}

    RangeDriftDetector._summary(
        out_of_range_series=out_of_range_series,
        out_of_range_series_ranges=out_of_range_series_ranges,
        out_of_range_exog=out_of_range_exog,
        out_of_range_exog_ranges=out_of_range_exog_ranges
    )

    captured = capsys.readouterr()
    assert "No series with out-of-range values found." in captured.out
    assert "'series_1': 'exog1' has values outside" in captured.out
    assert "'series_1': 'exog2' has values outside" in captured.out


def test_summary_both_out_of_range(capsys):
    """
    Test _summary with both out-of-range series and exogenous variables.
    """
    out_of_range_series = ["series1"]
    out_of_range_series_ranges = [(1.0, 10.0)]
    out_of_range_exog = ["exog1"]
    out_of_range_exog_ranges = [(0.0, 5.0)]

    RangeDriftDetector._summary(
        out_of_range_series=out_of_range_series,
        out_of_range_series_ranges=out_of_range_series_ranges,
        out_of_range_exog=out_of_range_exog,
        out_of_range_exog_ranges=out_of_range_exog_ranges
    )

    captured = capsys.readouterr()
    assert "'series1' has values outside the observed range [1.00000, 10.00000]." in captured.out
    assert "'exog1' has values outside the observed range [0.00000, 5.00000]." in captured.out



# Unit test _normalize_input
# ==============================================================================
import re
import pytest
import pandas as pd
from skforecast.drift_detection import RangeDriftDetector
from skforecast.exceptions import IgnoredArgumentWarning


def test_normalize_input_Series_without_name():
    """
    Test _normalize_input with a pandas Series without a name raises ValueError.
    """
    series = pd.Series([1, 2, 3])
    detector = RangeDriftDetector()

    msg = "test must have a name when a pandas Series is provided."
    with pytest.raises(ValueError, match=msg):
        detector._normalize_input(series, name="test")


def test_normalize_input_Series_with_name():
    """
    Test _normalize_input with a pandas Series that has a name.
    """
    series = pd.Series([1, 2, 3], name='y')
    detector = RangeDriftDetector()
    result = detector._normalize_input(series, name="series")
    expected = {'y': series}

    assert result == expected


def test_normalize_input_DataFrame_no_multiindex():
    """
    Test _normalize_input with a pandas DataFrame without MultiIndex.
    """
    df = pd.DataFrame({'col1': [1, 2], 'col2': [3, 4]})
    detector = RangeDriftDetector()
    result = detector._normalize_input(df, name="series")

    expected = {
        'col1': pd.Series([1, 2], name='col1'), 
        'col2': pd.Series([3, 4], name='col2')
    }
    
    assert result.keys() == expected.keys()
    for key in result:
        pd.testing.assert_series_equal(result[key], expected[key])


def test_normalize_input_DataFrame_no_multiindex_with_series_ids_and_series_specific_exog_():
    """
    Test _normalize_input with a pandas DataFrame without MultiIndex when
    series_ids is provided and series_specific_exog_ is True.
    """
    df = pd.DataFrame({'col1': [1, 2], 'col2': [3, 4]})
    detector = RangeDriftDetector()
    detector.series_specific_exog_ = True

    result = detector._normalize_input(
        df, name="series", series_ids=['col1', 'col2']
    )

    expected = {
        'col1': pd.DataFrame({'col1': [1, 2], 'col2': [3, 4]}),
        'col2': pd.DataFrame({'col1': [1, 2], 'col2': [3, 4]})
    }
    
    assert result.keys() == expected.keys()
    for key in result:
        pd.testing.assert_frame_equal(result[key], expected[key])


def test_normalize_input_DataFrame_multiindex_series():
    """
    Test _normalize_input with a pandas DataFrame with MultiIndex for 'series'.
    """
    index = pd.MultiIndex.from_tuples(
        [('series_1', 0), ('series_1', 1), ('series_2', 0), ('series_2', 1)],
        names=['series', 'time']
    )
    df = pd.DataFrame({'value': [1, 2, 3, 4]}, index=index)
    detector = RangeDriftDetector()
    result = detector._normalize_input(df, name="series")

    expected = {
        'series_1': pd.Series([1, 2], name='series_1'),
        'series_2': pd.Series([3, 4], name='series_2')
    }

    assert result.keys() == expected.keys()
    for key in result:
        pd.testing.assert_series_equal(result[key], expected[key], check_names=False)


def test_normalize_input_DataFrame_multiindex_last_window():
    """
    Test _normalize_input with a pandas DataFrame with MultiIndex for 'last_window'.
    """
    index = pd.MultiIndex.from_tuples(
        [('series_1', 0), ('series_1', 1), ('series_2', 0), ('series_2', 1)],
        names=['series', 'datetime']
    )
    df = pd.DataFrame({'value': [1, 2, 3, 4]}, index=index)
    detector = RangeDriftDetector()
    result = detector._normalize_input(df, name="last_window")

    expected = {
        'series_1': pd.Series([1, 2], name='series_1').rename_axis('datetime', axis='index'),
        'series_2': pd.Series([3, 4], name='series_2').rename_axis('datetime', axis='index')
    }

    assert result.keys() == expected.keys()
    for key in result:
        pd.testing.assert_series_equal(result[key], expected[key])


def test_normalize_input_DataFrame_multiindex_exog():
    """
    Test _normalize_input with a pandas DataFrame with MultiIndex for 'exog'.
    """
    index = pd.MultiIndex.from_tuples(
        [('series_1', 0), ('series_1', 1), ('series_2', 0), ('series_2', 1)],
        names=['series', 'datetime']
    )
    df = pd.DataFrame({'exog1': [1, 2, 3, 4], 'exog2': [5, 6, 7, 8]}, index=index)
    detector = RangeDriftDetector()
    result = detector._normalize_input(df, name="exog")

    expected = {
        'series_1': pd.DataFrame({'exog1': [1, 2], 'exog2': [5, 6]}, index=pd.Index([0, 1], name='datetime')),
        'series_2': pd.DataFrame({'exog1': [3, 4], 'exog2': [7, 8]}, index=pd.Index([0, 1], name='datetime'))
    }

    assert result.keys() == expected.keys()
    for key in result:
        pd.testing.assert_frame_equal(result[key], expected[key])


def test_normalize_warning_input_DataFrame_multiindex_series_multiple_columns():
    """
    Test _normalize_input with MultiIndex DataFrame for 'series' with multiple
    columns, should warn and use first.
    """
    index = pd.MultiIndex.from_tuples(
        [("series_1", 0), ("series_1", 1), ("series_2", 0), ("series_2", 1)],
        names=['series', 'datetime']
    )
    df = pd.DataFrame({'value1': [1, 2, 3, 4], 'value2': [5, 6, 7, 8]}, index=index)
    detector = RangeDriftDetector()

    msg = re.escape(
        "`series` DataFrame has multiple columns. Only the "
        "first column, 'value1', will be used. Others ignored.",
    )
    with pytest.warns(IgnoredArgumentWarning, match=msg):
        result = detector._normalize_input(df, name="series")
    
    expected = {
        'series_1': pd.Series([1, 2], name='series_1').rename_axis('datetime', axis='index'),
        'series_2': pd.Series([3, 4], name='series_2').rename_axis('datetime', axis='index')
    }

    assert result.keys() == expected.keys()
    for key in result:
        pd.testing.assert_series_equal(result[key], expected[key])


def test_normalize_input_dict_valid():
    """
    Test _normalize_input with a valid dict.
    """
    series1 = pd.Series([1, 2], name='s1')
    series2 = pd.Series([3, 4], name='s2')
    data = {'key1': series1, 'key2': series2}
    detector = RangeDriftDetector()
    result = detector._normalize_input(data, name="test")

    assert result == data


def test_normalize_raise_error_input_dict_invalid_value():
    """
    Test _normalize_input with a dict containing invalid value type.
    """
    data = {'key1': pd.Series([1, 2]), 'key2': [3, 4]}
    detector = RangeDriftDetector()

    msg = re.escape(
        "All values in `test` must be a pandas Series or DataFrame. " 
        "Review the value for key 'key2'."
    )
    with pytest.raises(TypeError, match=msg):
        detector._normalize_input(data, name="test")



# Unit test _create_lags ForecasterRecursiveMultiSeries
# ==============================================================================
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from ....preprocessing import RollingFeatures
from ....recursive import ForecasterRecursiveMultiSeries

rolling = RollingFeatures(
    stats=['mean', 'median', 'sum'], window_sizes=[5, 5, 6]
)


def test_create_lags_output():
    """
    Test matrix of lags is created properly when langs=3 and y=np.arange(10).
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3)
    results = forecaster._create_lags(y=np.arange(10))
    expected = (
        np.array([[2., 1., 0.],
                  [3., 2., 1.],
                  [4., 3., 2.],
                  [5., 4., 3.],
                  [6., 5., 4.],
                  [7., 6., 5.],
                  [8., 7., 6.]]),
        np.array([3., 4., 5., 6., 7., 8., 9.])
    )

    np.testing.assert_array_almost_equal(results[0], expected[0])
    np.testing.assert_array_almost_equal(results[1], expected[1])


def test_create_lags_output_interspersed_lags():
    """
    Test matrix of lags is is a list with interspersed lags.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=[4, 7])
    results = forecaster._create_lags(y=np.arange(10))
    expected = (
        np.array([[3., 0.],
                  [4., 1.],
                  [5., 2.]]),
        np.array([7., 8., 9.])
    )

    np.testing.assert_array_almost_equal(results[0], expected[0])
    np.testing.assert_array_almost_equal(results[1], expected[1])


def test_create_lags_output_pandas():
    """
    Test matrix of lags is created properly when X_as_pandas=True.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3)
    results = forecaster._create_lags(
        y=np.arange(10, dtype=float), X_as_pandas=True,
        train_index=pd.date_range('2020-01-03', periods=7, freq='D')
    )
    expected = (
        pd.DataFrame(
            data = np.array([
                       [2., 1., 0.],
                       [3., 2., 1.],
                       [4., 3., 2.],
                       [5., 4., 3.],
                       [6., 5., 4.],
                       [7., 6., 5.],
                       [8., 7., 6.]]
                   ),
            columns = ["lag_1", "lag_2", "lag_3"],
            index = pd.date_range('2020-01-03', periods=7, freq='D')
        ),
        np.array([3., 4., 5., 6., 7., 8., 9.])
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    np.testing.assert_array_almost_equal(results[1], expected[1])


def test_create_lags_when_window_size_window_features_greater_than_max_lag():
    """
    Test matrix of lags created properly when window_size of 
    window_features is greater than max lag.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, window_features=rolling
    )
    results = forecaster._create_lags(y=np.arange(10))
    expected = (
        np.array([[5., 4., 3.],
                  [6., 5., 4.],
                  [7., 6., 5.],
                  [8., 7., 6.]]),
        np.array([6., 7., 8., 9.])
    )

    np.testing.assert_array_almost_equal(results[0], expected[0])
    np.testing.assert_array_almost_equal(results[1], expected[1])


def test_create_lags_output_lags_None():
    """
    Test matrix of lags when lags=None.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=None, window_features=rolling
    )
    results = forecaster._create_lags(y=np.arange(10))
    expected = (
        None,
        np.array([6., 7., 8., 9.])
    )

    assert results[0] == expected[0]
    np.testing.assert_array_almost_equal(results[1], expected[1])



# Unit test symmetric_mean_absolute_percentage_error
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.metrics import symmetric_mean_absolute_percentage_error


def test_symmetric_mean_absolute_percentage_error_input_types():
    """
    Test input types of symmetric_mean_absolute_percentage_error.
    """
    y_true = np.array([1.0, 2.0, 3.0, 4.0])
    y_pred = np.array([1.1, 1.9, 3.2, 3.8])

    err_msg = re.escape("`y_true` must be a pandas Series or numpy ndarray")
    with pytest.raises(TypeError, match = err_msg):
        symmetric_mean_absolute_percentage_error([1, 2, 3], y_pred)
    
    err_msg = re.escape("`y_pred` must be a pandas Series or numpy ndarray")
    with pytest.raises(TypeError, match = err_msg):
        symmetric_mean_absolute_percentage_error(y_true, [1, 2, 3])


def test_symmetric_mean_absolute_percentage_error_input_length():
    """
    Test input lengths of symmetric_mean_absolute_percentage_error.
    """
    y_true = np.array([1.0, 2.0, 3.0, 4.0])
    y_pred = np.array([1.1, 1.9, 3.2])

    err_msg = re.escape("`y_true` and `y_pred` must have the same length")
    with pytest.raises(ValueError, match = err_msg):
        symmetric_mean_absolute_percentage_error(y_true, y_pred)

    err_msg = re.escape("`y_true` and `y_pred` must have the same length")
    with pytest.raises(ValueError, match = err_msg):
        symmetric_mean_absolute_percentage_error(y_true, y_pred)


def test_symmetric_mean_absolute_percentage_error_empty_input():
    """
    Test empty input of symmetric_mean_absolute_percentage_error.
    """
    y_true = np.array([])
    y_pred = np.array([])

    err_msg = re.escape("`y_true` and `y_pred` must have at least one element")
    with pytest.raises(ValueError, match = err_msg):
        symmetric_mean_absolute_percentage_error(y_true, y_pred)


def test_symmetric_mean_absolute_percentage_error_output():
    """
    Check that the output of symmetric_mean_absolute_percentage_error is correct.
    """
    y_true = np.array([30, 33.3, 38, 42, 31, 29.5, 43.5, 35.9, 37, 40])
    y_pred = np.array([34, 31, 43, 41.4, 35.6, 33, 40, 38, 34, 44.5])

    expected_smape = 9.162048875098746

    assert np.isclose(
        symmetric_mean_absolute_percentage_error(y_true, y_pred), 
        expected_smape
    )


def test_symmetric_mean_absolute_percentage_error_pandas_series_input():
    """
    Test pandas Series input of symmetric_mean_absolute_percentage_error.
    """
    y_true = pd.Series([30, 33.3, 38, 42, 31, 29.5, 43.5, 35.9, 37, 40])
    y_pred = pd.Series([34, 31, 43, 41.4, 35.6, 33, 40, 38, 34, 44.5])

    expected_smape = 9.162048875098746

    assert np.isclose(
        symmetric_mean_absolute_percentage_error(y_true, y_pred), 
        expected_smape
    )    



# Unit test _create_train_X_y ForecasterRnn using PyTorch backend
# ==============================================================================
import os
import re
import pytest
import numpy as np
import pandas as pd
os.environ["KERAS_BACKEND"] = "torch"
import keras
from sklearn.preprocessing import StandardScaler
from skforecast.deep_learning.utils import create_and_compile_model
from skforecast.deep_learning import ForecasterRnn

series = pd.DataFrame(np.random.randn(100, 3), columns=['l1', 'l2', 'l3'])


def test_create_train_X_y_TypeError_when_series_not_dataframe():
    """
    Test TypeError is raised when series is not a pandas DataFrame.
    """
    series = pd.DataFrame(np.random.randn(100, 3), columns=['l1', 'l2', 'l3'])
    model = create_and_compile_model(
                series=series, 
                levels="l1",    
                lags=3,           
                steps=1,              
                recurrent_layer="LSTM",
                recurrent_units=100,
                dense_units=[128, 64],
            )
    forecaster = ForecasterRnn(estimator=model, lags=3, levels="l1")
    series = pd.Series(np.arange(7))

    err_msg = (
        "`series` must be a pandas DataFrame. Got <class 'pandas.core.series.Series'>."
    )
    with pytest.raises(TypeError, match=err_msg):
        forecaster._create_train_X_y(series=series)


def test_create_train_X_y_ValueError_when_number_of_series_in_series_not_equal_to_forecaster():
    """
    Test ValueError is raised when number of series in `series` is not equal to
    the number of series expected by the forecaster.
    """
    series = pd.DataFrame(np.random.randn(100, 3), columns=['l1', 'l2', 'l3'])
    model = create_and_compile_model(
                series=series[['l1', 'l2']], 
                levels="l1",    
                lags=3,           
                steps=1,              
                recurrent_layer="LSTM",
                recurrent_units=100,
                dense_units=[128, 64],
            )
    forecaster = ForecasterRnn(estimator=model, lags=3, levels="l1")

    err_msg = re.escape(
        "Number of series in `series` (3) "
        "does not match the number of series expected by the model "
        "architecture (2)."
    )
    with pytest.raises(ValueError, match=err_msg):
        forecaster._create_train_X_y(series=series)


def test_create_train_X_y_ValueError_when_series_not_include_levels():
    """
    Test ValueError is raised when `levels` defined when initializing the forecaster
    is not included in `series` used for training.
    """
    series = pd.DataFrame(np.random.randn(100, 3), columns=['l1', 'l2', 'l3'])
    model = create_and_compile_model(
                series=series, 
                levels="l1",    
                lags=3,           
                steps=1,              
                recurrent_layer="LSTM",
                recurrent_units=100,
                dense_units=[128, 64],
            )
    forecaster = ForecasterRnn(estimator=model, lags=3, levels="l1")
    
    series_not_levels = series.copy()
    series_not_levels.columns = ['l4', 'l5', 'l6']

    err_msg = (
        f"`levels` defined when initializing the forecaster must be "
        f"included in `series` used for training. "
        f"{set(forecaster.levels) - set(list(series_not_levels.columns))} not found."
    )
    with pytest.raises(ValueError, match=err_msg):
        forecaster._create_train_X_y(series=series_not_levels)


def test_create_train_X_y_ValueError_when_len_series_is_lower_than_maximum_window_size_plus_steps():
    """
    Test ValueError is raised when length of series is lower than maximum window_size 
    plus number of steps included in the forecaster.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(5)),  
                           'l2': pd.Series(np.arange(5))})
    model = create_and_compile_model(
                series=series, 
                levels="l1",    
                lags=3,           
                steps=3,              
                recurrent_layer="LSTM",
                recurrent_units=100,
                dense_units=[128, 64],
            )
    forecaster = ForecasterRnn(estimator=model, lags=3, levels="l1")

    err_msg = re.escape(
        "Minimum length of `series` for training this forecaster is "
        "6. Reduce the number of "
        "predicted steps, 3, or the maximum "
        "lag, 3, if no more data is available.\n"
        "    Length `series`: 5.\n"
        "    Max step : 3.\n"
        "    Lags window size: 3."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster._create_train_X_y(series=series)


def test_create_train_X_y_ValueError_when_exog_not_included_but_model_requires_exog():
    """
    Test ValueError is raised when `exog` is not included but the model
    requires `exog` to be present.
    """
    series = pd.DataFrame(np.random.randn(100, 3), columns=['l1', 'l2', 'l3'])
    exog = pd.DataFrame(np.random.randn(100, 2), columns=['exog1', 'exog2'])
    model = create_and_compile_model(
                series=series,
                exog=exog, 
                levels="l1",    
                lags=3,           
                steps=1,              
                recurrent_layer="LSTM",
                recurrent_units=100,
                dense_units=[128, 64],
            )
    forecaster = ForecasterRnn(estimator=model, lags=3, levels="l1")

    err_msg = re.escape(
        "The estimator architecture expects exogenous variables during "
        "training. Please provide the `exog` argument. If this is "
        "unexpected, check your estimator architecture or the "
        "initialization parameters of the forecaster."
    )
    with pytest.raises(ValueError, match=err_msg):
        forecaster._create_train_X_y(series=series)


def test_create_train_X_y_ValueError_when_exog_included_but_model_no_exog():
    """
    Test ValueError is raised when `exog` is included but the model
    does not require `exog` to be present.
    """
    series = pd.DataFrame(np.random.randn(100, 3), columns=['l1', 'l2', 'l3'])
    exog = pd.DataFrame(np.random.randn(100, 2), columns=['exog1', 'exog2'])
    model = create_and_compile_model(
                series=series,
                levels="l1",    
                lags=3,           
                steps=1,              
                recurrent_layer="LSTM",
                recurrent_units=100,
                dense_units=[128, 64],
            )
    forecaster = ForecasterRnn(estimator=model, lags=3, levels="l1")

    err_msg = re.escape(
        "Exogenous variables (`exog`) were provided, but the model "
        "architecture was not built to expect exogenous variables. Please "
        "remove the `exog` argument or rebuild the model to include "
        "exogenous inputs."
    )
    with pytest.raises(ValueError, match=err_msg):
        forecaster._create_train_X_y(series=series, exog=exog)


def test_create_train_X_y_ValueError_when_number_of_exog_columns_not_equal_to_model():
    """
    Test ValueError is raised when number of columns in `exog` is not equal to
    the number of columns expected by the model.
    """
    series = pd.DataFrame(np.random.randn(100, 3), columns=['l1', 'l2', 'l3'])
    exog = pd.DataFrame(np.random.randn(100, 2), columns=['exog1', 'exog2'])
    model = create_and_compile_model(
                series=series,
                exog=exog, 
                levels="l1",    
                lags=3,    
                steps=1,              
                recurrent_layer="LSTM",
                recurrent_units=100,
                dense_units=[128, 64],
            )
    forecaster = ForecasterRnn(estimator=model, lags=3, levels="l1")

    err_msg = re.escape(
        "Number of columns in `exog` (1) "
        "does not match the number of exogenous variables expected "
        "by the model architecture (2)."
    )
    with pytest.raises(ValueError, match=err_msg):
        forecaster._create_train_X_y(series=series, exog=exog.iloc[:, :1])


@pytest.mark.parametrize("exog", 
    [pd.Series(np.arange(15), name='exog'), 
     pd.DataFrame(np.arange(50).reshape(25, 2)),
     pd.Series(np.arange(50), index=pd.date_range(start='2022-01-01', periods=50, freq='1D'), name='exog')
])
def test_create_train_X_y_ValueError_when_series_and_exog_have_different_length(exog):
    """
    Test ValueError is raised when length of series is not equal to length exog.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10)), 
                           'l2': pd.Series(np.arange(10))})
    if isinstance(exog.index, pd.DatetimeIndex):
        series.index = exog.index[:10]
    
    model = create_and_compile_model(
                series=series,
                exog=exog, 
                levels="l1",    
                lags=3,    
                steps=3,              
                recurrent_layer="LSTM",
                recurrent_units=100,
                dense_units=[128, 64],
            )
    forecaster = ForecasterRnn(estimator=model, lags=3, levels="l1")

    len_exog = len(exog)
    len_series = len(series)
    series_index_no_ws = series.index[forecaster.window_size:]
    len_series_no_ws = len(series_index_no_ws)
    err_msg = re.escape(
        f"Length of `exog` must be equal to the length of `series` (if "
        f"index is fully aligned) or length of `series` - `window_size` "
        f"(if `exog` starts after the first `window_size` values).\n"
        f"    `exog`                   : ({exog.index[0]} -- {exog.index[-1]})  (n={len_exog})\n"
        f"    `series`                 : ({series.index[0]} -- {series.index[-1]})  (n={len_series})\n"
        f"    `series` - `window_size` : ({series_index_no_ws[0]} -- {series_index_no_ws[-1]})  (n={len_series_no_ws})"
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster._create_train_X_y(series=series, exog=exog)


def test_create_train_X_y_ValueError_when_exog_columns_same_as_series_col_names():
    """
    Test ValueError is raised when an exog column is named the same as
    the series columns.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10)), 
                           'l2': pd.Series(np.arange(10))})
    exog = pd.DataFrame({'l1': pd.Series(np.arange(10)), 
                         'exog2': pd.Series(np.arange(10))})

    model = create_and_compile_model(
                series=series,
                exog=exog, 
                levels="l1",    
                lags=3,    
                steps=3,              
                recurrent_layer="LSTM",
                recurrent_units=100,
                dense_units=[128, 64],
            )
    forecaster = ForecasterRnn(estimator=model, lags=3, levels="l1")

    series_col_names = list(series.columns)
    exog_col_names = list(exog.columns)

    err_msg = re.escape(
        f"`exog` cannot contain a column named the same as one of "
        f"the series (column names of series).\n"
        f"  `series` columns : {series_col_names}.\n"
        f"  `exog`   columns : {exog_col_names}."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster._create_train_X_y(series=series, exog=exog)


def test_create_train_X_y_ValueError_when_transformer_exog_alter_shape_of_exog():
    """
    Test ValueError is raised when transformer_exog alters the shape of `exog`.
    """
    series = pd.DataFrame(
        {"l1": pd.Series(np.arange(10)), "l2": pd.Series(np.arange(10))}
    )
    series.index = pd.date_range(start="2022-01-01", periods=10, freq="D")
    exog = pd.DataFrame(
        {"exog_1": pd.Series(np.arange(10)), "exog_2": pd.Series(np.arange(10))}
    )
    exog.index = pd.date_range(start="2022-01-01", periods=10, freq="D")

    model = create_and_compile_model(
                series=series,
                exog=exog, 
                levels="l1",    
                lags=3,    
                steps=3,              
                recurrent_layer="LSTM",
                recurrent_units=100,
                dense_units=[128, 64],
            )
    
    class CustomTransformer:  # pragma: no cover
        """
        Custom transformer that alters the shape of `exog`.
        """
        def fit(self, X, y=None):
            return self

        def transform(self, X):
            self.categories_ = ['exog_1', 'exog_2', 'exog_3']
            return X.assign(exog_3=X['exog_1'] + X['exog_2'])
        
        def fit_transform(self, X, y=None):
            return self.transform(X)
    
    forecaster = ForecasterRnn(
        estimator=model, lags=3, levels="l1", transformer_exog=CustomTransformer()
    )

    err_msg = re.escape(
        "Number of columns in `exog` after transformation (3) "
        "does not match the number of columns before transformation (2). "
        "The ForecasterRnn does not support transformations that "
        "change the number of columns in `exog`. Preprocess `exog` "
        "before passing it to the `create_and_compile_model` function."
    )
    with pytest.raises(ValueError, match=err_msg):
        forecaster._create_train_X_y(series=series, exog=exog)


def test_create_train_X_y_ValueError_when_series_and_exog_have_different_index_but_same_length():
    """
    Test ValueError is raised when series and exog have different index but same length.
    """
    series = pd.DataFrame(
        {"l1": pd.Series(np.arange(10)), "l2": pd.Series(np.arange(10))}
    )
    series.index = pd.date_range(start="2022-01-01", periods=10, freq="1D")
    exog = pd.Series(
        np.arange(10), index=pd.RangeIndex(start=0, stop=10, step=1), name="exog"
    )

    model = create_and_compile_model(
                series=series,
                exog=exog, 
                levels="l1",    
                lags=3,    
                steps=3,              
                recurrent_layer="LSTM",
                recurrent_units=100,
                dense_units=[128, 64],
            )
    forecaster = ForecasterRnn(estimator=model, lags=3, levels="l1")

    err_msg = re.escape(
        "When `exog` has the same length as `series`, the index "
        "of `exog` must be aligned with the index of `series` "
        "to ensure the correct alignment of values."
    )
    with pytest.raises(ValueError, match=err_msg):
        forecaster._create_train_X_y(series=series, exog=exog)


def test_create_train_X_y_ValueError_when_series_and_exog_have_different_index_and_length_exog_no_window_size():
    """
    Test ValueError is raised when y and exog have different index and
    length exog no window_size.
    """
    series = pd.DataFrame(
        {"l1": pd.Series(np.arange(10)), "l2": pd.Series(np.arange(10))}
    )
    series.index = pd.date_range(start="2022-01-01", periods=10, freq="1D")
    exog = pd.Series(
        np.arange(3, 10), index=pd.RangeIndex(start=3, stop=10, step=1), name="exog"
    )

    model = create_and_compile_model(
                series=series,
                exog=exog, 
                levels="l1",    
                lags=3,    
                steps=3,              
                recurrent_layer="LSTM",
                recurrent_units=100,
                dense_units=[128, 64],
            )
    forecaster = ForecasterRnn(estimator=model, lags=3, levels="l1")

    err_msg = re.escape(
        "When `exog` doesn't contain the first `window_size` "
        "observations, the index of `exog` must be aligned with "
        "the index of `series` minus the first `window_size` "
        "observations to ensure the correct alignment of values."
    )
    with pytest.raises(ValueError, match=err_msg):
        forecaster._create_train_X_y(series=series, exog=exog)


def test_create_train_X_y_UserWarning_when_levels_of_transformer_series_not_equal_to_series_col_names():
    """
    Test UserWarning is raised when `transformer_series` is a dict and its keys are
    not the same as forecaster.series_col_names.
    """
    series = pd.DataFrame({"1": pd.Series(np.arange(5)), "2": pd.Series(np.arange(5))})
    dict_transformers = {"1": StandardScaler(), "3": StandardScaler()}
    model = create_and_compile_model(
                series=series, 
                levels="1",    
                lags=3,           
                steps=1,              
                recurrent_layer="LSTM",
                recurrent_units=100,
                dense_units=[128, 64],
            )
    forecaster = ForecasterRnn(
        estimator=model, levels="1", transformer_series=dict_transformers, lags=3
    )

    series_not_in_transformer_series = set(series.columns) - set(
        forecaster.transformer_series.keys()
    )

    warn_msg = re.escape(
        f"{series_not_in_transformer_series} not present in `transformer_series`."
        f" No transformation is applied to these series."
    )
    with pytest.warns(UserWarning, match=warn_msg):
        forecaster._create_train_X_y(series=series)


def test_create_train_X_y_ValueError_when_all_series_values_are_missing():
    """
    Test ValueError is raised when all series values are missing.
    """
    series = pd.DataFrame({"1": pd.Series(np.arange(7)), "2": pd.Series([np.nan] * 7)})
    series.index = pd.date_range(start="2022-01-01", periods=7, freq="1D")
    model = create_and_compile_model(
                series=series, 
                levels="1",    
                lags=3,           
                steps=1,              
                recurrent_layer="LSTM",
                recurrent_units=100,
                dense_units=[128, 64],
            )
    forecaster = ForecasterRnn(estimator=model, levels="1", lags=3)

    err_msg = re.escape("`y` has missing values.")
    with pytest.raises(ValueError, match=err_msg):
        forecaster._create_train_X_y(series=series)


@pytest.mark.parametrize(
    "values",
    [
        [0, 1, 2, 3, 4, 5, np.nan],
        [0, 1] + [np.nan] * 5,
        [np.nan, 1, 2, 3, 4, 5, np.nan],
        [0, 1, np.nan, 3, np.nan, 5, 6],
        [np.nan, np.nan, np.nan, 3, np.nan, 5, 6],
    ],
)
def test_create_train_X_y_ValueError_when_series_values_are_missing(values):
    """
    Test ValueError is raised when series values are missing in different
    locations.
    """
    series = pd.DataFrame({"1": pd.Series(values), "2": pd.Series(np.arange(7))})
    series.index = pd.date_range(start="2022-01-01", periods=7, freq="1D")
    model = create_and_compile_model(
                series=series, 
                levels="1",    
                lags=3,           
                steps=1,              
                recurrent_layer="LSTM",
                recurrent_units=100,
                dense_units=[128, 64],
            )
    forecaster = ForecasterRnn(estimator=model, levels="1", lags=3)

    err_msg = re.escape("`y` has missing values.")
    with pytest.raises(ValueError, match=err_msg):
        forecaster._create_train_X_y(series=series)


def test_create_train_X_y_output_when_series_10_and_transformer_series_is_StandardScaler():
    """
    Test the output of create_train_X_y when exog is None and transformer_series
    is StandardScaler.
    """
    series = pd.DataFrame(
        {
            "l1": pd.Series(np.arange(10, dtype=float)),
            "l2": pd.Series(np.arange(10, dtype=float)),
        }
    )
    model = create_and_compile_model(
                series=series, 
                levels="l1",    
                lags=5,           
                steps=2,              
                recurrent_layer="LSTM",
                recurrent_units=128,
                dense_units=64,
            )
    forecaster = ForecasterRnn(
        estimator=model, levels="l1", transformer_series=StandardScaler(), lags=5
    )

    (
        X_train,
        exog_train,
        y_train,
        dimension_names,
        exog_names_in_,
        exog_dtypes_in_,
        exog_dtypes_out_
    ) = forecaster._create_train_X_y(series=series)

    expected_X_train = np.array(
        [[[-1.5666989 , -1.5666989 ],
        [-1.21854359, -1.21854359],
        [-0.87038828, -0.87038828],
        [-0.52223297, -0.52223297],
        [-0.17407766, -0.17407766]],

       [[-1.21854359, -1.21854359],
        [-0.87038828, -0.87038828],
        [-0.52223297, -0.52223297],
        [-0.17407766, -0.17407766],
        [ 0.17407766,  0.17407766]],

       [[-0.87038828, -0.87038828],
        [-0.52223297, -0.52223297],
        [-0.17407766, -0.17407766],
        [ 0.17407766,  0.17407766],
        [ 0.52223297,  0.52223297]],

       [[-0.52223297, -0.52223297],
        [-0.17407766, -0.17407766],
        [ 0.17407766,  0.17407766],
        [ 0.52223297,  0.52223297],
        [ 0.87038828,  0.87038828]]]
    )
    expected_exog_train = None
    expected_y_train = np.array(
        [
            [[0.17407766], [0.52223297]],
            [[0.52223297], [0.87038828]],
            [[0.87038828], [1.21854359]],
            [[1.21854359], [1.5666989]],
        ]
    )
    expected_exog_names_in_ = None
    expected_exog_dtypes_in_ = None
    expected_exog_dtypes_out_ = None

    expected_dimension_names = {
        "X_train": {
            0: pd.RangeIndex(start=5, stop=9, step=1),
            1: ['lag_5', 'lag_4', 'lag_3', 'lag_2', 'lag_1'],
            2: ["l1", "l2"],
        },
        "y_train": {
            0: pd.RangeIndex(start=5, stop=9, step=1),
            1: ["step_1", "step_2"],
            2: ["l1"]
        },
        "exog_train": {0: None, 1: None, 2: None}
    }
    
    np.testing.assert_almost_equal(X_train, expected_X_train)
    assert exog_train is expected_exog_train
    np.testing.assert_almost_equal(y_train, expected_y_train)
    assert exog_names_in_ is expected_exog_names_in_
    assert exog_dtypes_in_ is expected_exog_dtypes_in_
    assert exog_dtypes_out_ is expected_exog_dtypes_out_
    assert dimension_names['X_train'][0].equals(expected_dimension_names['X_train'][0])
    assert dimension_names['X_train'][1] == expected_dimension_names['X_train'][1]
    assert dimension_names['X_train'][2] == expected_dimension_names['X_train'][2]
    assert dimension_names['y_train'][0].equals(expected_dimension_names['y_train'][0])
    assert dimension_names['y_train'][1] == expected_dimension_names['y_train'][1]
    assert dimension_names['y_train'][2] == expected_dimension_names['y_train'][2]
    assert dimension_names['exog_train'] == expected_dimension_names['exog_train']


def test_create_train_X_y_output_when_series_10_and_transformer_series_is_StandardScaler_and_exog():
    """
    Test the output of create_train_X_y when exog is None and transformer_series
    is StandardScaler and exog is provided.
    """
    series = pd.DataFrame(
        {
            "l1": pd.Series(np.arange(10, dtype=float)),
            "l2": pd.Series(np.arange(10, dtype=float)),
        }
    )
    exog = pd.DataFrame(
        {
            "exog1": pd.Series(np.arange(10, dtype=float)),
            "exog2": pd.Series(np.arange(10, dtype=float)),
        }
    )
    model = create_and_compile_model(
                series=series,
                exog=exog,
                levels="l1",
                lags=5,
                steps=2,
                recurrent_layer="LSTM",
                recurrent_units=128,
                dense_units=64,
            )
    forecaster = ForecasterRnn(
        estimator=model, levels="l1", transformer_series=StandardScaler(), lags=5
    )

    (
        X_train,
        exog_train,
        y_train,
        dimension_names,
        exog_names_in_,
        exog_dtypes_in_,
        exog_dtypes_out_
    ) = forecaster._create_train_X_y(series=series, exog=exog)

    expected_X_train = np.array(
        [[[-1.5666989 , -1.5666989 ],
        [-1.21854359, -1.21854359],
        [-0.87038828, -0.87038828],
        [-0.52223297, -0.52223297],
        [-0.17407766, -0.17407766]],

       [[-1.21854359, -1.21854359],
        [-0.87038828, -0.87038828],
        [-0.52223297, -0.52223297],
        [-0.17407766, -0.17407766],
        [ 0.17407766,  0.17407766]],

       [[-0.87038828, -0.87038828],
        [-0.52223297, -0.52223297],
        [-0.17407766, -0.17407766],
        [ 0.17407766,  0.17407766],
        [ 0.52223297,  0.52223297]],

       [[-0.52223297, -0.52223297],
        [-0.17407766, -0.17407766],
        [ 0.17407766,  0.17407766],
        [ 0.52223297,  0.52223297],
        [ 0.87038828,  0.87038828]]]
    )
    expected_exog_train = np.array(
        [[[0.55555556, 0.55555556],
        [0.66666667, 0.66666667]],
       [[0.66666667, 0.66666667],
        [0.77777778, 0.77777778]],
       [[0.77777778, 0.77777778],
        [0.88888889, 0.88888889]],
       [[0.88888889, 0.88888889],
        [1.        , 1.        ]]]
    )
    expected_y_train = np.array(
        [
            [[0.17407766], [0.52223297]],
            [[0.52223297], [0.87038828]],
            [[0.87038828], [1.21854359]],
            [[1.21854359], [1.5666989]],
        ]
    )
    expected_exog_names_in_ = ['exog1', 'exog2']
    expected_exog_dtypes_in_ = {'exog1': np.dtype('float64'), 'exog2': np.dtype('float64')}
    expected_exog_dtypes_out_ = {'exog1': np.dtype('float64'), 'exog2': np.dtype('float64')}

    expected_dimension_names = {
        "X_train": {
            0: pd.RangeIndex(start=5, stop=9, step=1),
            1: ['lag_5', 'lag_4', 'lag_3', 'lag_2', 'lag_1'],
            2: ["l1", "l2"],
        },
        "y_train": {
            0: pd.RangeIndex(start=5, stop=9, step=1),
            1: ["step_1", "step_2"],
            2: ["l1"]
        },
        "exog_train": {
            0: pd.RangeIndex(start=5, stop=9, step=1),
            1: ['step_1', 'step_2'],
            2: ['exog1', 'exog2']
        }
    }
    
    np.testing.assert_almost_equal(X_train, expected_X_train)
    np.testing.assert_almost_equal(exog_train, expected_exog_train)
    np.testing.assert_almost_equal(y_train, expected_y_train)
    assert exog_names_in_ == expected_exog_names_in_
    assert exog_dtypes_in_ == expected_exog_dtypes_in_
    assert exog_dtypes_out_ == expected_exog_dtypes_out_
    assert dimension_names['X_train'][0].equals(expected_dimension_names['X_train'][0])
    assert dimension_names['X_train'][1] == expected_dimension_names['X_train'][1]
    assert dimension_names['X_train'][2] == expected_dimension_names['X_train'][2]
    assert dimension_names['y_train'][0].equals(expected_dimension_names['y_train'][0])
    assert dimension_names['y_train'][1] == expected_dimension_names['y_train'][1]
    assert dimension_names['y_train'][2] == expected_dimension_names['y_train'][2]
    assert dimension_names['exog_train'][0].equals(expected_dimension_names['exog_train'][0])
    assert dimension_names['exog_train'][1] == expected_dimension_names['exog_train'][1]
    assert dimension_names['exog_train'][2] == expected_dimension_names['exog_train'][2]



# Unit test save_forecaster and load_forecaster
# ==============================================================================
import os
import re
import joblib
import pytest
import inspect
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import StandardScaler

from .... import __version__
from ....recursive import ForecasterRecursive
from ....recursive import ForecasterRecursiveMultiSeries
from ...utils import save_forecaster
from ...utils import load_forecaster
from ....exceptions import SkforecastVersionWarning, SaveLoadSkforecastWarning


def custom_weights(y):  # pragma: no cover
    """
    """
    return np.ones(len(y))


def custom_weights2(y):  # pragma: no cover
    """
    """
    return np.arange(1, len(y) + 1)


class UserWindowFeature:  # pragma: no cover
    def __init__(self, window_sizes, features_names):
        self.window_sizes = window_sizes
        self.features_names = features_names

    def transform_batch(self):
        pass

    def transform(self):
        pass


def test_save_and_load_forecaster_persistence():
    """ 
    Test if a loaded forecaster is exactly the same as the original one.
    """
    forecaster = ForecasterRecursive(
        estimator=LinearRegression(), lags=3, transformer_y=StandardScaler()
    )
    rng = np.random.default_rng(12345)
    y = pd.Series(rng.normal(size=100))
    forecaster.fit(y=y)
    save_forecaster(forecaster=forecaster, file_name='forecaster.joblib', verbose=True)
    forecaster_loaded = load_forecaster(file_name='forecaster.joblib', verbose=True)
    os.remove('forecaster.joblib')

    for key in vars(forecaster).keys():
    
        attribute_forecaster = forecaster.__getattribute__(key)
        attribute_forecaster_loaded = forecaster_loaded.__getattribute__(key)

        if key in ['estimator', 'binner', 'transformer_y', 'transformer_exog']:
            assert joblib.hash(attribute_forecaster) == joblib.hash(attribute_forecaster_loaded)
        elif isinstance(attribute_forecaster, np.ndarray):
            np.testing.assert_array_almost_equal(attribute_forecaster, attribute_forecaster_loaded)
        elif isinstance(attribute_forecaster, pd.Series):
            pd.testing.assert_series_equal(attribute_forecaster, attribute_forecaster_loaded)
        elif isinstance(attribute_forecaster, pd.DataFrame):
            pd.testing.assert_frame_equal(attribute_forecaster, attribute_forecaster_loaded)
        elif isinstance(attribute_forecaster, pd.Index):
            pd.testing.assert_index_equal(attribute_forecaster, attribute_forecaster_loaded)
        elif isinstance(attribute_forecaster, dict):
            assert attribute_forecaster.keys() == attribute_forecaster_loaded.keys()
            for k in attribute_forecaster.keys():
                if isinstance(attribute_forecaster[k], np.ndarray):
                    np.testing.assert_array_almost_equal(attribute_forecaster[k], attribute_forecaster_loaded[k])
                elif isinstance(attribute_forecaster[k], pd.Series):
                    pd.testing.assert_series_equal(attribute_forecaster[k], attribute_forecaster_loaded[k])
                elif isinstance(attribute_forecaster[k], pd.DataFrame):
                    pd.testing.assert_frame_equal(attribute_forecaster[k], attribute_forecaster_loaded[k])
                elif isinstance(attribute_forecaster[k], pd.Index):
                    pd.testing.assert_index_equal(attribute_forecaster[k], attribute_forecaster_loaded[k])
                else:
                    assert attribute_forecaster[k] == attribute_forecaster_loaded[k]
        else:
            assert attribute_forecaster == attribute_forecaster_loaded


def test_save_and_load_forecaster_SkforecastVersionWarning():
    """ 
    Test warning used to notify that the skforecast version installed in the 
    environment differs from the version used to create the forecaster.
    """
    forecaster = ForecasterRecursive(estimator=LinearRegression(), lags=3)
    rng = np.random.default_rng(123)
    y = pd.Series(rng.normal(size=100))
    forecaster.fit(y=y)
    forecaster.skforecast_version = '0.0.0'
    save_forecaster(forecaster=forecaster, file_name='forecaster.joblib', verbose=False)

    warn_msg = re.escape(
        f"The skforecast version installed in the environment differs "
        f"from the version used to create the forecaster.\n"
        f"    Installed Version  : {__version__}\n"
        f"    Forecaster Version : 0.0.0\n"
        f"This may create incompatibilities when using the library."
    )
    with pytest.warns(SkforecastVersionWarning, match = warn_msg):
        load_forecaster(file_name='forecaster.joblib', verbose=False)
        os.remove('forecaster.joblib')


@pytest.mark.parametrize("weight_func", 
                         [custom_weights, 
                          {'serie_1': custom_weights, 
                           'serie_2': custom_weights2}, 
                          {'serie_1': custom_weights}], 
                         ids = lambda func: f'type: {type(func)}')
def test_save_forecaster_save_custom_functions(weight_func):
    """ 
    Test if custom functions are saved correctly.
    """
    series = pd.DataFrame(
        {'serie_1': np.random.normal(size=20),
         'serie_2': np.random.normal(size=20)}
    ).to_dict(orient='series')
    
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LinearRegression(),
                     lags               = 5,
                     weight_func        = weight_func,
                     transformer_series = StandardScaler()
                 )
    forecaster.fit(series=series)
    save_forecaster(
        forecaster=forecaster, file_name='forecaster.joblib', save_custom_functions=True
    )
    load_forecaster(file_name='forecaster.joblib', verbose=True)
    os.remove('forecaster.joblib')
    
    weight_functions = weight_func.values() if isinstance(weight_func, dict) else [weight_func]
    for weight_func in weight_functions:
        weight_func_file = weight_func.__name__ + '.py'
        assert os.path.exists(weight_func_file)
        with open(weight_func_file, 'r') as file:
            assert inspect.getsource(weight_func) == file.read()
        os.remove(weight_func_file)


@pytest.mark.parametrize("weight_func", 
                         [custom_weights, 
                          {'serie_1': custom_weights, 
                           'serie_2': custom_weights2}], 
                         ids = lambda func: f'func: {func}')
def test_save_forecaster_warning_dont_save_custom_functions(weight_func):
    """ 
    Test SaveLoadSkforecastWarning when custom functions are not saved.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator   = LinearRegression(),
                     lags        = 5,
                     weight_func = weight_func
                 )

    warn_msg = re.escape(
        "Custom function(s) used to create weights are not saved. "
        "To save them, set `save_custom_functions` to `True`."
    )
    with pytest.warns(SaveLoadSkforecastWarning, match = warn_msg):
        save_forecaster(
            forecaster=forecaster, file_name='forecaster.joblib', save_custom_functions=False
        )
        os.remove('forecaster.joblib')


def test_save_forecaster_warning_when_user_defined_window_features():
    """ 
    Test SaveLoadSkforecastWarning when user-defined window features.
    """

    window_features = UserWindowFeature(
        window_sizes=[1, 2], features_names=['feature_1', 'feature_2']
    )
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator       = LinearRegression(),
                     lags            = 5,
                     window_features = window_features
                 )

    warn_msg = re.escape(
        "The Forecaster includes custom user-defined classes in the "
        "`window_features` argument. These classes are not saved automatically "
        "when saving the Forecaster. Please ensure you save these classes "
        "manually and import them before loading the Forecaster.\n"
        "    Custom classes: " + ', '.join({'UserWindowFeature'}),
    )
    with pytest.warns(SaveLoadSkforecastWarning, match = warn_msg):
        save_forecaster(
            forecaster=forecaster, file_name='forecaster.joblib', save_custom_functions=False
        )
        os.remove('forecaster.joblib')



################################################################################
#                             RangeDriftDetector                               #
#                                                                              #
# This work by skforecast team is licensed under the BSD 3-Clause License.     #
################################################################################
# coding=utf-8

from __future__ import annotations
import pandas as pd
import warnings
import textwrap
from rich.console import Console
from rich.panel import Panel

from .. import __version__
from ..exceptions import (
    FeatureOutOfRangeWarning,
    IgnoredArgumentWarning,
    MissingExogWarning,
    UnknownLevelWarning
)
from ..utils import (
    set_skforecast_warnings,
    get_style_repr_html
)


class RangeDriftDetector:
    """
    Detector of out-of-range values based on training feature ranges.

    The detector is intentionally lightweight: it does not compute advanced
    drift statistics since it is used to check single observations during
    inference. Suitable for real-time applications.

    Parameters
    ----------
    self

    Attributes
    ----------
    series_names_in_ : list
        Names of the series used during training.
    series_values_range_ : dict
        Range of values of the target series used during training.
    exog_names_in_ : list
        Names of the exogenous variables used during training.
    exog_values_range_ : dict
        Range of values of the exogenous variables used during training.
    series_specific_exog_ : bool
        Indicates whether exogenous variables have different values across
        target series during training (i.e., exogenous is series-specific
        rather than global).
    is_fitted : bool
        Whether the detector has been fitted to the training data.
    
    """

    def __init__(self) -> None:

        self.series_names_in_      = None
        self.series_values_range_  = None
        self.exog_names_in_        = None
        self.exog_values_range_    = None
        self.series_specific_exog_ = False
        self.is_fitted             = False

    def __repr__(self) -> str:
        """
        Information displayed when a RangeDriftDetector object is printed.
        """
    
        series_names_in_ = None
        if self.series_names_in_ is not None:
            if len(self.series_names_in_) > 50:
                series_names_in_ = self.series_names_in_[:25] + ["..."] + self.series_names_in_[-25:]
                series_names_in_ = ", ".join(series_names_in_)
            else:
                series_names_in_ = ", ".join(self.series_names_in_)

        exog_names_in_ = None
        if self.exog_names_in_ is not None:
            if len(self.exog_names_in_) > 50:
                exog_names_in_ = self.exog_names_in_[:25] + ["..."] + self.exog_names_in_[-25:]
                exog_names_in_ = ", ".join(exog_names_in_)
            else:
                exog_names_in_ = ", ".join(self.exog_names_in_)

        info = (
            f"{'=' * len(type(self).__name__)} \n"
            f"{type(self).__name__} \n"
            f"{'=' * len(type(self).__name__)} \n"
            f"Fitted series          = {series_names_in_} \n"
            f"Series value ranges    = {self.series_values_range_} \n"
            f"Fitted exogenous       = {exog_names_in_} \n"
            f"Exogenous value ranges = {self.exog_values_range_} \n"
            f"Series-specific exog   = {self.series_specific_exog_} \n"
            f"Is fitted              = {self.is_fitted}"
        )

        return info

    def _repr_html_(self):
        """
        HTML representation of the object.
        The "General Information" section is expanded by default.
        """
    
        series_names_in_ = None
        if self.series_names_in_ is not None:
            if len(self.series_names_in_) > 50:
                series_names_in_ = self.series_names_in_[:25] + ["..."] + self.series_names_in_[-25:]
                series_names_in_ = ", ".join(series_names_in_)
            else:
                series_names_in_ = ", ".join(self.series_names_in_)

        exog_names_in_ = None
        if self.exog_names_in_ is not None:
            if len(self.exog_names_in_) > 50:
                exog_names_in_ = self.exog_names_in_[:25] + ["..."] + self.exog_names_in_[-25:]
                exog_names_in_ = ", ".join(exog_names_in_)
            else:
                exog_names_in_ = ", ".join(self.exog_names_in_)

        style, unique_id = get_style_repr_html(self.is_fitted)
        content = f"""
        <div class="container-{unique_id}">
            <p style="font-size: 1.5em; font-weight: bold; margin-block-start: 0.83em; margin-block-end: 0.83em;">{type(self).__name__}</p>
            <details open>
                <summary>General Information</summary>
                <ul>
                    <li><strong>Fitted series:</strong> {series_names_in_}</li>
                    <li><strong>Fitted exogenous:</strong> {exog_names_in_}</li>
                    <li><strong>Series-specific exogenous:</strong> {self.series_specific_exog_}</li>
                    <li><strong>Is fitted:</strong> {self.is_fitted}</li>
                </ul>
            </details>
            <details>
                <summary>Series value ranges</summary>
                <ul>
                    {self.series_values_range_}
                </ul>
            </details>
            <details>
                <summary>Exogenous value ranges</summary>
                <ul>
                    {self.exog_values_range_}
                </ul>
            </details>
            <p>
                <a href="https://skforecast.org/{__version__}/api/drift_detection.html#skforecast.drift_detection._range_drift.RangeDriftDetector">&#128712 <strong>API Reference</strong></a>
                &nbsp;&nbsp;
                <a href="https://skforecast.org/{__version__}/user_guides/drift-detection.html">&#128462 <strong>User Guide</strong></a>
            </p>
        </div>
        """
        
        return style + content

    @classmethod
    def _get_features_range(
        cls, 
        X: pd.Series | pd.DataFrame
    ) -> tuple | set | dict[str, tuple | set]:
        """
        Get a summary of the features in the DataFrame or Series. For numeric
        features, it returns the min and max values. For categorical features,
        it returns the unique values.

        Parameters
        ----------
        X : pandas Series, pandas DataFrame
            Input data to summarize.
        
        Returns
        -------
        features_ranges: tuple, set, dict
            Feature ranges. If X is a Series, returns a tuple (min, max) for numeric
            data or a set of unique values for categorical data. If X is a DataFrame,
            returns a dictionary with column names as keys and their respective ranges
            (tuple or set) as values.

        """

        if not isinstance(X, (pd.DataFrame, pd.Series)):
            raise TypeError("Input must be a pandas DataFrame or Series.")

        if isinstance(X, pd.Series):
            if pd.api.types.is_numeric_dtype(X):
                features_ranges = (float(X.min()), float(X.max()))
            else:
                features_ranges = set(X.dropna().unique())

        if isinstance(X, pd.DataFrame):
            num_cols = [
                col for col in X.columns if pd.api.types.is_numeric_dtype(X[col])
            ]
            cat_cols = [col for col in X.columns if col not in num_cols]

            features_ranges = {}
            features_ranges.update(
                {col: (float(X[col].min()), float(X[col].max())) for col in num_cols}
            )
            features_ranges.update(
                {col: set(X[col].dropna().unique()) for col in cat_cols}
            )

        return features_ranges

    @classmethod
    def _check_feature_range(
        cls,
        feature_range: tuple | set,
        X: pd.Series
    ) -> bool:
        """
        Check if there is any value outside the training range. For numeric features,
        it checks if the values are within the min and max range. For categorical features,
        it checks if the values are among the seen categories.

        Parameters
        ----------
        feature_range : tuple, set
            Output from _get_features_range() for a single feature.
        X : pd.Series
            New data to validate

        Returns
        -------
        bool
            True if there is any value outside the training range, False otherwise.
        
        """

        if isinstance(feature_range, tuple):
            return X.min() < feature_range[0] or X.max() > feature_range[1]
        else:
            unseen = set(X.dropna().unique()) - feature_range
            return bool(unseen)

    @classmethod
    def _display_warnings(
        cls,
        not_compliant_feature: str,
        feature_range: tuple | set,
        series_name: str = None,
    ) -> None:
        """
        Display warnings for features with values outside the training range.

        Parameters
        ----------
        not_compliant_feature : str
            Name of the feature with values outside the training range.
        feature_range : tuple | set
            Training range of the feature.
        series_name : str, optional
            Name of the series being checked, if applicable.

        Returns
        -------
        None

        """

        if isinstance(feature_range, tuple):
            # Numeric
            msg = (
                f"'{not_compliant_feature}' has values outside the range seen during training "
                f"[{feature_range[0]:.5f}, {feature_range[1]:.5f}]. "
                f"This may affect the accuracy of the predictions."
            )
        else:
            # Categorical
            msg = (
                f"'{not_compliant_feature}' has values not seen during training. Seen values: "
                f"{feature_range}. This may affect the accuracy of the predictions."
            )

        if series_name:
            msg = f"'{series_name}': " + msg

        warnings.warn(msg, FeatureOutOfRangeWarning)

    @classmethod
    def _summary(
        cls,
        out_of_range_series: list,
        out_of_range_series_ranges: list,
        out_of_range_exog: list,
        out_of_range_exog_ranges: list
    ) -> None:
        """
        Summarize the results of the range check.

        Parameters
        ----------
        out_of_range_series : list
            List of series names that are out of range.
        out_of_range_series_ranges : list
            List of ranges for the out-of-range series.
        out_of_range_exog : list
            List of exogenous variable names that are out of range.
        out_of_range_exog_ranges : list
            List of ranges for the out-of-range exogenous variables.

        Returns
        -------
        None

        """
        
        msg_series = ""
        if out_of_range_series:
            series_msgs = []
            for series, series_range in zip(
                out_of_range_series, out_of_range_series_ranges
            ):
                msg_temp = (
                    f"'{series}' has values outside the observed range "
                    f"[{series_range[0]:.5f}, {series_range[1]:.5f}]."
                )
                series_msgs.append(textwrap.fill(msg_temp, width=80))
            msg_series = "\n".join(series_msgs) + "\n"
        else:
            msg_series = "No series with out-of-range values found.\n"

        msg_exog = ""
        if out_of_range_exog:
            exog_msgs = []
            if isinstance(out_of_range_exog, list):
                for exog, exog_range in zip(out_of_range_exog, out_of_range_exog_ranges):
                    if isinstance(exog_range, tuple):
                        # Numeric
                        msg_temp = (
                            f"'{exog}' has values outside the observed range "
                            f"[{exog_range[0]:.5f}, {exog_range[1]:.5f}]."
                        )
                    else:
                        # Categorical
                        msg_temp = (
                            f"'{exog}' has values not seen during training. Seen values: "
                            f"{exog_range}."
                        )
                    exog_msgs.append(textwrap.fill(msg_temp, width=80))
            else:
                for key, value in out_of_range_exog.items():
                    for exog, exog_range in zip(value, out_of_range_exog_ranges[key]):
                        if isinstance(exog_range, tuple):
                            # Numeric
                            msg_temp = (
                                f"'{exog}' has values outside the observed range "
                                f"[{exog_range[0]:.5f}, {exog_range[1]:.5f}]."
                            )
                        else:
                            # Categorical
                            msg_temp = (
                                f"'{exog}' has values not seen during training. Seen values: "
                                f"{exog_range}."
                            )
                        msg_temp = f"'{key}': " + msg_temp
                        exog_msgs.append(textwrap.fill(msg_temp, width=80))

            msg_exog = "\n".join(exog_msgs)
        else:
            msg_exog = "No exogenous variables with out-of-range values found."

        console = Console()
        content = (
            f"[bold]Series:[/bold]\n{msg_series}\n"
            f"[bold]Exogenous Variables:[/bold]\n{msg_exog}"
        )
        console.print(Panel(content, title="[bold]Out-of-range summary[/bold]", expand=False))

    def _normalize_input(
        self, 
        X: pd.Series | pd.DataFrame | dict[str, pd.Series | pd.DataFrame],
        name: str,
        series_ids: list[str] | None = None
    ) -> dict[str, pd.Series | pd.DataFrame]:
        """
        Convert pd.Series, pd.DataFrame or dict into a standardized dict of
        pd.Series or pd.DataFrames.

        Parameters
        ----------
        X : pandas Series, pandas DataFrame, dict
            Input data to normalize.
        name : str
            Name of the input being normalized. Used for error messages.
            Expected values are 'series', 'last_window' or 'exog'.
        series_ids : list, default None
            Series IDs to include in the normalization of exogenous variables.

        Returns
        -------
        X : dict
            Normalized input as a dictionary of pandas Series or DataFrames.

        """

        if isinstance(X, pd.Series):
            if not X.name:
                raise ValueError(
                    f"{name} must have a name when a pandas Series is provided."
                )
            X = {X.name: X}

        elif isinstance(X, pd.DataFrame):
            if isinstance(X.index, pd.MultiIndex):
                if name in ["series", "last_window"]:
                    col = X.columns[0]
                    if len(X.columns) != 1:
                        warnings.warn(
                            f"`{name}` DataFrame has multiple columns. Only the "
                            f"first column, '{col}', will be used. Others ignored.",
                            IgnoredArgumentWarning,
                        )
                    X = {
                        series_id: X.loc[series_id][col].rename(series_id)
                        for series_id in X.index.levels[0]
                    }
                else:
                    X = {series_id: X.loc[series_id] for series_id in X.index.levels[0]}
            else:
                if self.series_specific_exog_ and series_ids:
                    X = {series_id: X.copy() for series_id in series_ids}
                else:
                    X = X.to_dict(orient="series")

        elif isinstance(X, dict):
            for k, v in X.items():
                if not isinstance(v, (pd.Series, pd.DataFrame)):
                    raise TypeError(
                        f"All values in `{name}` must be a pandas Series or DataFrame. "
                        f"Review the value for key '{k}'."
                    )

        return X
    
    def fit(
        self,
        series: pd.DataFrame | pd.Series | dict[str, pd.Series | pd.DataFrame] | None = None,
        exog: pd.DataFrame | pd.Series | dict[str, pd.Series | pd.DataFrame] | None = None,
        **kwargs
    ) -> None:
        """
        Fit detector, storing training ranges.

        Parameters
        ----------
        series : pandas Series, pandas DataFrame, dict, aliases: `y`
            Input time series data to fit the detector, ideally the same ones
            used to fit the forecaster.
        exog : pandas Series, pandas DataFrame, dict, default None
            Exogenous variables to include in the forecaster.

        Returns
        -------
        None

        """

        if series is None and ('y' not in kwargs or kwargs['y'] is None):
            raise ValueError(
                "One of `series` or `y` must be provided."
            )
        if 'y' in kwargs:
            if series is not None:
                raise ValueError(
                    "Cannot specify both `series` and `y`. Please provide only one of them."
                )
            series = kwargs.pop('y')

        if not isinstance(series, (pd.Series, pd.DataFrame, dict)):
            raise TypeError("Input must be a pandas Series, DataFrame or dict.")

        if not isinstance(exog, (pd.Series, pd.DataFrame, dict, type(None))):
            raise TypeError(
                "Exogenous variables must be a pandas Series, DataFrame or dict."
            )
        
        self.series_names_in_      = []
        self.series_values_range_  = {}
        self.exog_names_in_        = None
        self.exog_values_range_    = None
        self.series_specific_exog_ = False
        self.is_fitted             = False

        series = self._normalize_input(series, name="series")
        for key, value in series.items():
            self.series_names_in_.append(key)
            self.series_values_range_[key] = self._get_features_range(X=value)

        if exog is not None:

            exog = self._normalize_input(exog, name="exog")

            self.exog_names_in_ = []
            self.exog_values_range_ = {}
            for key, value in exog.items():
                if isinstance(value, pd.Series):
                    self.exog_names_in_.append(key)
                else:
                    self.exog_names_in_.extend(value.columns)
                self.exog_values_range_[key] = self._get_features_range(X=value)

            self.exog_names_in_ = list(dict.fromkeys(self.exog_names_in_))
            self.series_specific_exog_ = any(key in self.series_names_in_ for key in exog.keys())

        self.is_fitted = True

    def predict(
        self,
        last_window: pd.Series | pd.DataFrame | dict[str, pd.Series | pd.DataFrame] | None = None,
        exog: pd.Series | pd.DataFrame | dict[str, pd.Series | pd.DataFrame] | None = None,
        verbose: bool = True,
        suppress_warnings: bool = False
    ) -> tuple[bool, list[str], list[str] | dict[str, list[str]]]:
        """
        Check if there is any value outside the training range for last_window and exog.

        Parameters
        ----------
        last_window : pandas Series, pandas DataFrame, dict, default None
            Series values used to create the predictors (lags) needed in the
            first iteration of the prediction (t + 1).
        exog : pandas Series, pandas DataFrame, dict, default None
            Exogenous variable/s included as predictor/s.
        verbose : bool, default False
            Whether to print a summary of the check.
        suppress_warnings : bool, default False
            Whether to suppress warnings.

        Returns
        -------
        flag_out_of_range : bool
            True if there is any value outside the training range, False otherwise.
        out_of_range_series : list
            List of series names that are out of range.
        out_of_range_exog : list, dict
            Exogenous variables that are out of range.

            - If `self.series_specific_exog_` is False: returns a list with the names
            of exogenous variables that are out of range (global exogenous).
            - If `self.series_specific_exog_` is True: returns a dictionary where
            keys are series names and values are lists of out-of-range exogenous
            variables for each series.

        """

        if not self.is_fitted:
            raise RuntimeError("Model is not fitted yet.")

        if not isinstance(last_window, (pd.Series, pd.DataFrame, dict, type(None))):
            raise TypeError(
                "`last_window` must be a pandas Series, DataFrame, dict or None."
            )

        if not isinstance(exog, (pd.Series, pd.DataFrame, dict, type(None))):
            raise TypeError(
                "`exog` must be a pandas Series, DataFrame, dict or None."
            )
        
        set_skforecast_warnings(suppress_warnings, action='ignore')
        
        flag_out_of_range = False

        out_of_range_series = []
        out_of_range_series_ranges = []
        if last_window is not None:
            last_window = self._normalize_input(last_window, name="last_window")
            for key, value in last_window.items():
                if isinstance(value, pd.Series):
                    value = value.to_frame()
                for col in value.columns:
                    if key not in self.series_names_in_:
                        warnings.warn(
                            f"'{key}' was not seen during training. Its range is unknown.",
                            UnknownLevelWarning
                        )
                        continue
                    is_out_of_range = self._check_feature_range(
                        feature_range=self.series_values_range_[col], X=value[col]
                    )
                    if is_out_of_range:
                        flag_out_of_range = True
                        out_of_range_series.append(col)
                        out_of_range_series_ranges.append(self.series_values_range_[col])
                        self._display_warnings(
                            not_compliant_feature = col,
                            feature_range         = self.series_values_range_[col],
                            series_name           = None
                        )

        out_of_range_exog = {} if self.series_specific_exog_ else []
        out_of_range_exog_ranges = {} if self.series_specific_exog_ else []
        if exog is not None:
            series_ids = list(last_window.keys()) if last_window is not None else self.series_names_in_
            exog = self._normalize_input(exog, name="exog", series_ids=series_ids)
            for key, value in exog.items():

                if isinstance(value, pd.Series):
                    value = value.to_frame()
                features_ranges = self.exog_values_range_.get(key, None)
                
                if self.series_specific_exog_:
                    out_of_range_exog[key] = []
                    out_of_range_exog_ranges[key] = []
                
                for col in value.columns:

                    if not isinstance(features_ranges, dict):
                        features_ranges = {key: features_ranges}
                    
                    if col not in self.exog_names_in_:
                        warnings.warn(
                            f"'{col}' was not seen during training. Its range is unknown.",
                            MissingExogWarning,
                        )
                        continue

                    is_out_of_range = self._check_feature_range(
                        feature_range=features_ranges[col], X=value[col]
                    )
                    
                    if is_out_of_range:

                        flag_out_of_range = True
                        if self.series_specific_exog_:
                            out_of_range_exog[key].append(col)
                            out_of_range_exog_ranges[key].append(features_ranges[col])
                        else:
                            out_of_range_exog.append(col)
                            out_of_range_exog_ranges.append(features_ranges[col])

                        self._display_warnings(
                            not_compliant_feature = col,
                            feature_range         = features_ranges[col],
                            series_name           = key if self.series_specific_exog_ else None,
                        )

                if self.series_specific_exog_ and not out_of_range_exog[key]:
                    out_of_range_exog.pop(key)
                    out_of_range_exog_ranges.pop(key)

        if verbose:
            self._summary(
                out_of_range_series        = out_of_range_series,
                out_of_range_series_ranges = out_of_range_series_ranges,
                out_of_range_exog          = out_of_range_exog,
                out_of_range_exog_ranges   = out_of_range_exog_ranges
            )

        set_skforecast_warnings(suppress_warnings, action='default')

        return flag_out_of_range, out_of_range_series, out_of_range_exog



# Unit test _check_interval
# ==============================================================================
import re
import pytest
from skforecast.utils.utils import check_interval


def test_check_interval_TypeError_when_interval_is_not_a_list():
    """
    Check `TypeError` is raised when `interval` is not a `list`.
    """
    err_msg = re.escape(
        "`interval` must be a `list` or `tuple`. For example, interval of 95% "
        "should be as `interval = [2.5, 97.5]`."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_interval(interval = 'not_a_list')


def test_check_interval_ValueError_when_interval_len_is_not_2():
    """
    Check `ValueError` is raised when `interval` len is not 2.
    """
    err_msg = re.escape(
        "`interval` must contain exactly 2 values, respectively the "
        "lower and upper interval bounds. For example, interval of 95% "
        "should be as `interval = [2.5, 97.5]`."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_interval(interval = [2.5, 50.0, 97.5])


def test_check_interval_ValueError_when_interval_lower_bound_less_than_0():
    """
    Check `ValueError` is raised when lower bound is less than 0.
    """
    err_msg = re.escape("Lower interval bound (-1.0) must be >= 0 and < 100.")
    with pytest.raises(ValueError, match = err_msg):
        check_interval(interval = [-1.0, 97.5])


@pytest.mark.parametrize("interval", 
                         [[100.0, 97.5], (101.0, 97.5)], 
                         ids = lambda value: f'interval: {value}')
def test_check_interval_ValueError_when_interval_lower_bound_greater_than_or_equal_to_100(interval):
    """
    Check `ValueError` is raised when lower bound is greater than or equal to 100.
    """
    err_msg = re.escape(f"Lower interval bound ({interval[0]}) must be >= 0 and < 100.")
    with pytest.raises(ValueError, match = err_msg):
        check_interval(interval = interval)


@pytest.mark.parametrize("interval", 
                         [[2.5, 0.0], (2.5, -1.0)], 
                         ids = lambda value: f'interval: {value}')
def test_check_interval_ValueError_when_interval_upper_bound_less_than_or_equal_to_0(interval):
    """
    Check `ValueError` is raised when upper bound is less than or equal to 0.
    """
    err_msg = re.escape(f"Upper interval bound ({interval[1]}) must be > 0 and <= 100.")
    with pytest.raises(ValueError, match = err_msg):
        check_interval(interval = interval)


def test_check_interval_ValueError_when_interval_upper_bound_greater_than_100():
    """
    Check `ValueError` is raised when upper bound is greater than 100.
    """
    err_msg = re.escape('Upper interval bound (101.0) must be > 0 and <= 100.')
    with pytest.raises(ValueError, match = err_msg):
        check_interval(interval = [2.5, 101.0])


@pytest.mark.parametrize("interval", 
                         [[2.5, 2.5], (2.5, 2.0)], 
                         ids = lambda value: f'interval: {value}')
def test_check_interval_ValueError_when_interval_lower_bound_greater_than_or_equal_to_upper_bound(interval):
    """
    Check `ValueError` is raised when lower bound is greater than or equal to
    upper bound.
    """
    err_msg = re.escape(
        f"Lower interval bound ({interval[0]}) must be less than the "
        f"upper interval bound ({interval[1]})."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_interval(interval = interval)


@pytest.mark.parametrize("interval", 
                         [[0, 95], (20, 81)], 
                         ids = lambda value: f'interval: {value}')
def test_check_interval_ValueError_when_interval_is_not_symmetric(interval):
    """
    Check `ValueError` is raised when interval is not symmetric.
    """
    err_msg = re.escape(
        f"Interval must be symmetric, the sum of the lower, ({interval[0]}), "
        f"and upper, ({interval[1]}), interval bounds must be equal to "
        f"100. Got {interval[0] + interval[1]}."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_interval(interval = interval, ensure_symmetric_intervals = True)


def test_check_interval_TypeError_when_quantiles_is_not_a_list():
    """
    Check `TypeError` is raised when `quantiles` is not a `list`.
    """
    err_msg = re.escape(
        "`quantiles` must be a `list` or `tuple`. For example, quantiles "
        "0.05, 0.5, and 0.95 should be as `quantiles = [0.05, 0.5, 0.95]`."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_interval(quantiles = 'not_a_list')


@pytest.mark.parametrize("quantiles", 
                         [[-0.01, 0.01, 0.5], [0., 1., 1.1], [-2], (-2, 2)], 
                         ids = lambda value: f'quantiles: {value}')
def test_check_interval_ValueError_when_elements_in_quantiles_are_out_of_bounds(quantiles):
    """
    Check `ValueError` is raised when any element in `quantiles` is 
    not between 0 and 100.
    """
    err_msg = re.escape("All elements in `quantiles` must be >= 0 and <= 1.")
    with pytest.raises(ValueError, match = err_msg):
        check_interval(quantiles=quantiles)


def test_check_interval_TypeError_when_alpha_is_not_float():
    """
    Check `TypeError` is raised when `alpha` is not a `float`.
    """
    err_msg = re.escape(
        "`alpha` must be a `float`. For example, interval of 95% "
        "should be as `alpha = 0.05`."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_interval(alpha = 'not_a_float')


@pytest.mark.parametrize("alpha", 
                         [1., 0.], 
                         ids = lambda value: f'alpha: {value}')
def test_check_interval_ValueError_when_alpha_is_out_of_bounds(alpha):
    """
    Check `ValueError` is raised when alpha is not between 0 and 1.
    """
    err_msg = re.escape(f'`interval` must have a value between 0 and 1. Got {alpha}.')
    with pytest.raises(ValueError, match = err_msg):
        check_interval(alpha=alpha, alpha_literal='interval')




# Unit test __init__ ForecasterRnn using PyTorch backend
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
import os
os.environ["KERAS_BACKEND"] = "torch"
import keras
from skforecast.deep_learning import ForecasterRnn
from skforecast.deep_learning import create_and_compile_model


def test_ForecasterRnn_init_basic_and_attr_types():
    """
    Basic initialization of ForecasterRnn with attributes types.
    """
    series = pd.DataFrame({'A': np.arange(10, dtype=float)})
    model = create_and_compile_model(
        series=series, lags=3, steps=2, model_name="modelo"
    )
    forecaster = ForecasterRnn(
        estimator=model,
        levels='A',
        lags=3,
        transformer_series=None,
        transformer_exog=None,
        fit_kwargs={},
        forecaster_id="f-id"
    )

    assert isinstance(forecaster.estimator, keras.Model)
    assert forecaster.levels == ['A']
    assert forecaster.transformer_series is None
    assert forecaster.transformer_exog is None
    assert forecaster.is_fitted is False
    assert isinstance(forecaster.creation_date, str)
    assert forecaster.fit_date is None
    assert forecaster.keras_backend_ is None
    assert isinstance(forecaster.skforecast_version, str)
    assert isinstance(forecaster.python_version, str)
    assert forecaster.forecaster_id == "f-id"
    assert forecaster._probabilistic_mode == "no_binned"
    assert forecaster.dropna_from_series is False
    assert forecaster.encoding is None
    assert forecaster.differentiation is None
    assert forecaster.layers_names == [
        'series_input', 'lstm_1', 'dense_1', 'output_dense_td_layer', 'reshape'
    ]
    np.testing.assert_array_almost_equal(forecaster.lags, np.array([1, 2, 3]))
    assert forecaster.window_size == forecaster.max_lag
    np.testing.assert_array_almost_equal(forecaster.steps, np.array([1, 2]))
    assert forecaster.max_step == 2
    assert forecaster.n_series_in == 1
    assert forecaster.n_levels_out == 1
    assert forecaster.exog_in_ is False
    assert forecaster.n_exog_in is None
    assert forecaster.series_val is None
    assert forecaster.exog_val is None
    assert isinstance(forecaster.fit_kwargs, dict)


def test_ForecasterRnn_init_levels_list_and_lags_list():
    """
    Supports levels and lags as lists.
    """
    series = pd.DataFrame({'A': np.arange(10), 'B': np.arange(10, 20)})
    model = create_and_compile_model(series=series, lags=[1, 2], steps=2)
    forecaster = ForecasterRnn(
        estimator=model,
        levels=['A', 'B'],
        lags=[1, 2]
    )

    assert forecaster.levels == ['A', 'B']
    np.testing.assert_array_almost_equal(forecaster.lags, np.array([1, 2]))
    assert forecaster.n_series_in == 2
    assert forecaster.n_levels_out == 2


def test_ForecasterRnn_init_raises_ValueError_if_number_of_lags_not_matching_estimator():
    """
    Raise ValueError if number of lags does not match the model.
    """
    series = pd.DataFrame({'A': np.arange(10)})
    model = create_and_compile_model(series=series, lags=2, steps=2)
    err_msg = re.escape(
        "Number of lags (3) does not match the number of "
        "lags expected by the estimator architecture (2)."
    )
    with pytest.raises(ValueError, match=err_msg):
        ForecasterRnn(
            estimator=model,
            levels='A',
            lags=3
        )


def test_ForecasterRnn_init_raises_TypeError_if_levels_not_str_or_list():
    """
    Raise TypeError if levels is not str or list.
    """
    series = pd.DataFrame({'A': np.arange(10)})
    model = create_and_compile_model(series=series, lags=3, steps=2)
    err_msg = re.escape(
        "`levels` argument must be a string or list. Got <class 'int'>."
    )
    with pytest.raises(TypeError, match=err_msg):
        ForecasterRnn(
            estimator=model,
            levels=7,
            lags=3
        )


def test_ForecasterRnn_init_raises_ValueError_if_levels_list_length_not_matching_estimator():
    """
    Raise ValueError if levels list length does not match the model.
    """
    series = pd.DataFrame({'A': np.arange(10)})
    model = create_and_compile_model(series=series, lags=3, steps=2, levels=['A'])
    err_msg = re.escape(
        "Number of levels (3) does not match the number of "
        "levels expected by the estimator architecture (1)."
    )
    with pytest.raises(ValueError, match=err_msg):
        ForecasterRnn(
            estimator=model,
            levels=['A', 'B', 'C'],
            lags=3
        )


def test_ForecasterRnn_init_with_exog_and_check_exog_attrs():
    """
    Initialization with model that has exogenous variables: related attributes.
    """
    series = pd.DataFrame({'A': np.arange(10)})
    exog = pd.DataFrame({'exog1': np.arange(10)})
    model = create_and_compile_model(series=series, lags=3, steps=2, exog=exog)
    forecaster = ForecasterRnn(
        estimator=model,
        levels='A',
        lags=3
    )
    assert forecaster.exog_in_ is True
    assert forecaster.n_exog_in == 1
    assert "exog_input" in forecaster.layers_names
    assert forecaster.n_levels_out == 1


def test_ForecasterRnn_init_with_series_val_and_exog_val():
    """
    Initialization with fit_kwargs: series_val and exog_val.
    """
    series = pd.DataFrame({'A': np.arange(10)})
    exog = pd.DataFrame({'exog1': np.arange(10)})
    model = create_and_compile_model(series=series, lags=3, steps=2, exog=exog)
    series_val = pd.DataFrame({'A': np.arange(3)})
    exog_val = pd.DataFrame({'exog1': np.arange(3)})
    fit_kwargs = {
        "series_val": series_val,
        "exog_val": exog_val
    }
    forecaster = ForecasterRnn(
        estimator=model,
        levels='A',
        lags=3,
        fit_kwargs=fit_kwargs
    )
    assert isinstance(forecaster.series_val, pd.DataFrame)
    assert isinstance(forecaster.exog_val, pd.DataFrame)
    assert "series_val" not in forecaster.fit_kwargs and "exog_val" not in forecaster.fit_kwargs


def test_ForecasterRnn_init_raises_TypeError_if_series_val_not_dataframe():
    """
    Raise TypeError if series_val is not DataFrame.
    """
    series = pd.DataFrame({'A': np.arange(10)})
    exog = pd.DataFrame({'exog1': np.arange(10)})
    model = create_and_compile_model(series=series, lags=3, steps=2, exog=exog)
    fit_kwargs = {
        "series_val": [1, 2, 3],
        "exog_val": pd.DataFrame({'exog1': np.arange(3)})
    }
    err_msg = re.escape(
        "`series_val` must be a pandas DataFrame. Got <class 'list'>."
    )
    with pytest.raises(TypeError, match=err_msg):
        ForecasterRnn(
            estimator=model,
            levels='A',
            lags=3,
            fit_kwargs=fit_kwargs
        )


def test_ForecasterRnn_init_raises_ValueError_if_exog_in_but_missing_exog_val():
    """
    Raise ValueError if there is series_val but missing exog_val and the model uses exogenous variables.
    """
    series = pd.DataFrame({'A': np.arange(10)})
    exog = pd.DataFrame({'exog1': np.arange(10)})
    model = create_and_compile_model(series=series, lags=3, steps=2, exog=exog)
    fit_kwargs = {
        "series_val": pd.DataFrame({'A': np.arange(3)})
        # No exog_val
    }
    err_msg = re.escape(
        "If `series_val` is provided, `exog_val` must also be "
        "provided using the `fit_kwargs` argument when the "
        "estimator has exogenous variables."
    )
    with pytest.raises(ValueError, match=err_msg):
        ForecasterRnn(
            estimator=model,
            levels='A',
            lags=3,
            fit_kwargs=fit_kwargs
        )


def test_ForecasterRnn_init_raises_TypeError_if_exog_val_wrong_type():
    """
    Raise TypeError if exog_val is not DataFrame or Series.
    """
    series = pd.DataFrame({'A': np.arange(10)})
    exog = pd.DataFrame({'exog1': np.arange(10)})
    model = create_and_compile_model(series=series, lags=3, steps=2, exog=exog)
    fit_kwargs = {
        "series_val": pd.DataFrame({'A': np.arange(3)}),
        "exog_val": 42
    }
    err_msg = re.escape(
        "`exog_val` must be a pandas Series or DataFrame. Got <class 'int'>."
    )
    with pytest.raises(TypeError, match=err_msg):
        ForecasterRnn(
            estimator=model,
            levels='A',
            lags=3,
            fit_kwargs=fit_kwargs
        )



# Unit test _create_data_to_return_dict ForecasterDirectMultiVariate
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from skforecast.preprocessing import RollingFeatures
from skforecast.direct import ForecasterDirectMultiVariate

rolling = RollingFeatures(
    stats=['mean', 'median', 'sum'], window_sizes=[5, 5, 6]
)


def test_create_data_to_return_dict_ValueError_when_lags_keys_not_in_series():
    """
    Test ValueError is raised when lags keys are not in series column names.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10)),  
                           'l2': pd.Series(np.arange(10))})
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', steps=2, lags={'l1': 2, '2': 3}
    )
    
    series_names_in_ = list(series.columns)

    err_msg = re.escape(
        f"When `lags` parameter is a `dict`, its keys must be the "
        f"same as `series` column names. If don't want to include lags, "
         "add '{column: None}' to the lags dict.\n" 
        f"  Lags keys        : {list(forecaster.lags.keys())}.\n"
        f"  `series` columns : {series_names_in_}."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster._create_data_to_return_dict(series_names_in_=series_names_in_)

  
def test_create_data_to_return_dict_when_same_lags_all_series():
    """
    Test _create_data_to_return_dict output when lags is the same for all series.
    """    
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', steps=1, lags=3
    )
    
    results = forecaster._create_data_to_return_dict(series_names_in_=['l1', 'l2'])
    expected = (
        {'l1': 'both', 'l2': 'X'},
        ['l1', 'l2']
    )

    for k in forecaster.lags_.keys():
        np.testing.assert_array_almost_equal(forecaster.lags_[k], np.array([1, 2, 3]))
    for k in forecaster.lags_names.keys():
        assert forecaster.lags_names[k] == [f'{k}_lag_{i}' for i in range(1, 4)]
    for k in results[0].keys():
        assert results[0][k] == expected[0][k]
    assert results[1] == expected[1]

  
def test_create_data_to_return_dict_when_interspersed_lags_all_series():
    """
    Test _create_data_to_return_dict output when interspersed lags for all series.
    """    
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', steps=1, lags=[1, 5]
    )
    
    results = forecaster._create_data_to_return_dict(series_names_in_=['l1', 'l2'])
    expected = (
        {'l1': 'both', 'l2': 'X'},
        ['l1', 'l2']
    )

    for k in forecaster.lags_.keys():
        np.testing.assert_array_almost_equal(forecaster.lags_[k], np.array([1, 5]))
    for k in forecaster.lags_names.keys():
        assert forecaster.lags_names[k] == [f'{k}_lag_1', f'{k}_lag_5']
    for k in results[0].keys():
        assert results[0][k] == expected[0][k]
    assert results[1] == expected[1]

  
def test_create_data_to_return_dict_when_lags_dict_with_all_series():
    """
    Test _create_data_to_return_dict output when lags is a dict with lags for 
    all series.
    """    
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l2', steps=1, lags={'l1': 3, 'l2': 4}
    )
    
    results = forecaster._create_data_to_return_dict(series_names_in_=['l1', 'l2'])
    expected = (
        {'l1': 'X', 'l2': 'both'},
        ['l1', 'l2']
    )

    for k in forecaster.lags_.keys():
        expected_lags = [1, 2, 3] if k == 'l1' else [1, 2, 3, 4]
        np.testing.assert_array_almost_equal(forecaster.lags_[k], np.array(expected_lags))
    for k in forecaster.lags_names.keys():
        range_ = range(1, 4) if k == 'l1' else range(1, 5)
        expected_names = [f'{k}_lag_{i}' for i in range_]
        assert forecaster.lags_names[k] == expected_names
    for k in results[0].keys():
        assert results[0][k] == expected[0][k]
    assert results[1] == expected[1]


def test_create_data_to_return_dict_when_lags_dict_with_no_lags_for_level():
    """
    Test _create_data_to_return_dict output when lags is a dict with no lags for
    the level.
    """    
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l2', steps=1, lags={'l1': 3, 'l2': None}
    )
    
    results = forecaster._create_data_to_return_dict(series_names_in_=['l1', 'l2'])
    expected = (
        {'l1': 'X', 'l2': 'y'},
        ['l1']
    )

    for k in forecaster.lags_.keys():
        if forecaster.lags_[k] is not None:
            np.testing.assert_array_almost_equal(forecaster.lags_[k], np.array([1, 2, 3]))
        else:
            assert forecaster.lags_[k] is None
    for k in forecaster.lags_names.keys():
        if forecaster.lags_names[k] is not None:
            assert forecaster.lags_names[k] == [f'{k}_lag_{i}' for i in range(1, 4)]
        else:
            assert forecaster.lags_names[k] is None
    for k in results[0].keys():
        assert results[0][k] == expected[0][k]
    assert results[1] == expected[1]


def test_create_data_to_return_dict_when_lags_dict_with_lags_only_for_level():
    """
    Test _create_data_to_return_dict output when lags is a dict with lags only for
    level.
    """    
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', steps=1, lags={'l1': 3, 'l2': None}
    )
    
    results = forecaster._create_data_to_return_dict(series_names_in_=['l1', 'l2'])
    expected = (
        {'l1': 'both'},
        ['l1']
    )

    for k in forecaster.lags_.keys():
        if forecaster.lags_[k] is not None:
            np.testing.assert_array_almost_equal(forecaster.lags_[k], np.array([1, 2, 3]))
        else:
            assert forecaster.lags_[k] is None
    for k in forecaster.lags_names.keys():
        if forecaster.lags_names[k] is not None:
            assert forecaster.lags_names[k] == [f'{k}_lag_{i}' for i in range(1, 4)]
        else:
            assert forecaster.lags_names[k] is None
    for k in results[0].keys():
        assert results[0][k] == expected[0][k]
    assert results[1] == expected[1]


def test_create_data_to_return_dict_when_lags_dict_with_lags_only_for_level_and_window_features():
    """
    Test _create_data_to_return_dict output when lags is a dict with lags only for
    level and window_features.
    """    
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', steps=1, lags={'l1': 3, 'l2': None}, window_features=rolling
    )
    
    results = forecaster._create_data_to_return_dict(series_names_in_=['l1', 'l2'])
    expected = (
        {'l1': 'both'},
        ['l1', 'l2']
    )

    for k in forecaster.lags_.keys():
        if forecaster.lags_[k] is not None:
            np.testing.assert_array_almost_equal(forecaster.lags_[k], np.array([1, 2, 3]))
        else:
            assert forecaster.lags_[k] is None
    for k in forecaster.lags_names.keys():
        if forecaster.lags_names[k] is not None:
            assert forecaster.lags_names[k] == [f'{k}_lag_{i}' for i in range(1, 4)]
        else:
            assert forecaster.lags_names[k] is None
    for k in results[0].keys():
        assert results[0][k] == expected[0][k]
    assert results[1] == expected[1]


def test_create_data_to_return_dict_when_lags_is_None():
    """
    Test _create_data_to_return_dict output when lags is None.
    """    
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', steps=1, lags=None, window_features=rolling
    )
    
    results = forecaster._create_data_to_return_dict(series_names_in_=['l1', 'l2'])
    expected = (
        {'l1': 'y'},
        ['l1', 'l2']
    )

    for k in forecaster.lags_.keys():
        assert forecaster.lags_[k] is None
    for k in forecaster.lags_names.keys():
        assert forecaster.lags_names[k] is None
    for k in results[0].keys():
        assert results[0][k] == expected[0][k]
    assert results[1] == expected[1]
